import numpy as np
import sys
import os
import subprocess as sp
from PIL import Image, ImageDraw
from scipy.misc import imsave
from scipy.signal import medfilt2d
import ast
import json

from fit_ellipse import fit_ellipse, FitEllipse
from itracker_utils import generate_rays, initial_pupil_point, initial_cr_point, sobel_grad
import logging

import matplotlib.pyplot as plt

# import cv2

color_list = ['b','g','r','c','m','y','k']

class iTracker (object):
    def __init__(self, output_folder, 
                 im_shape, num_frames, 
                 input_stream, 
                 threshold_factor=1.3, auto=True,
                 cutoff_pixels=10,
                 bbox_pupil=None,
                 bbox_cr=None):

        self.im_shape = im_shape
        self.num_frames = num_frames
        self.movie_shape = (num_frames, im_shape[0], im_shape[1])
        self.input_stream = input_stream

        self.threshold_factor = threshold_factor
        self.auto = auto
        self.folder = output_folder
        self.cutoff_pixels = cutoff_pixels
        self.bbox_pupil = bbox_pupil
        self.bbox_cr = bbox_cr

        self._mean_frame = None

        if not os.path.exists(self.folder):
            os.mkdir(self.folder)

        self.run_params_file = os.path.join(self.folder, 'run_params.json')
        self.run_params = { 'threshold_factor': threshold_factor,
                            'auto':  auto,
                            'cutoff_pixels': cutoff_pixels,
                            'movie_shape':  self.movie_shape,
                            'im_shape':  im_shape,
                            'bbox_pupil':  bbox_pupil,
                            'bbox_cr':  bbox_cr }
        with open(self.run_params_file, 'w') as f:
            f.write(json.dumps(self.run_params))

        self.movie_path_storage_file = os.path.join(self.folder, 'movie_path.txt')
        if os.path.exists(self.movie_path_storage_file):
            with open(self.movie_path_storage_file, 'r') as f:
                self.movie_path = f.read()
        else:
            self.movie_path = None

        self.input_image_folder = os.path.join(self.folder, 'input_images')
        if not os.path.exists(self.input_image_folder):
            os.mkdir(self.input_image_folder)

        self.results_folder = os.path.join(self.folder,'results')

        if not os.path.exists(self.results_folder):
            os.mkdir(self.results_folder)

        # self.rays_folder = os.path.join(self.results_folder,'rays')
        # if not os.path.exists(self.rays_folder):
        #     os.mkdir(self.rays_folder)

        self.qc_folder = os.path.join(self.results_folder,'qc')
        if not os.path.exists(self.qc_folder):
            os.mkdir(self.qc_folder)

        self.frames_folder = os.path.join(self.results_folder,'output_frames')
        if not os.path.exists(self.frames_folder):
            os.mkdir(self.frames_folder)

        self.pupil_file = os.path.join(self.results_folder, 'pupil_params.npy')
        self.cr_file = os.path.join(self.results_folder, 'cr_params.npy')
        self.mean_frame_file = os.path.join(self.results_folder, 'mean_frame.npy')
        self.annotated_movie_file = os.path.join(self.results_folder, 'annotated_movie.mp4')

        # add variables to determine whether to provide diagnostic, QC and other output
        # method to regnerate image frames, with or without results?


    # fix this so it sets an absolute path
    def set_movie(self, file_path):
        logging.debug("Setting movie_path to: %s", file_path)
        self.movie_path = file_path
        with open(self.movie_path_storage_file, 'w') as f:
            f.write(self.movie_path)

    def set_bbox_pupil(self, bbox):
        self.bbox_pupil = bbox
        self.run_params['bbox_pupil']=self.bbox_pupil
        with open(self.run_params_file, 'w') as f:
            f.write(json.dumps(self.run_params))

    def set_bbox_cr(self, bbox):
        self.bbox_cr = bbox
        self.run_params['bbox_cr']=self.bbox_cr
        with open(self.run_params_file, 'w') as f:
            f.write(json.dumps(self.run_params))

    def create_input_images(self, image_type='png'):
        self.input_stream.create_images(self.input_image_folder, image_type)

    @property
    def mean_frame(self):
        if self._mean_frame is None:
            self._mean_frame = self.compute_mean_frame()
        return self._mean_frame

    @mean_frame.setter
    def mean_frame(self, mean_frame):
        self._mean_frame = mean_frame

    def estimate_bbox_from_mean_frame(self, margin=75, image_type='png'):
        try:
            import keras
        except ImportError:
            logging.debug("keras failed to import.  Returning None for bbox_pupil and bbox_cr")
            return None, None

        from keras.applications import InceptionV3

        logging.debug("Estimating bbox parameters from 'mean_frame'")
        # compute the representation for mean_frame
        model = InceptionV3(include_top=False, weights='imagenet')
        # print self.mean_frame.dtype, self.mean_frame.shape
        mp_temp = self.mean_frame.astype(np.float32)
        mp_temp -= 128
        mp_temp /= 128
        rep = model.predict(mp_temp.reshape((1,)+mp_temp.shape))  # shape (1,13,18,2048)
        rep[rep<0]=0
        rep = np.mean(rep, axis=(0,1,2))  # shape (2048,)

        # load regression weights
        module_folder = os.path.dirname(os.path.abspath(__file__))
        W_pupil = np.load(os.path.join(module_folder,'resources','pupil_weights.npy'))  # shape (2048, 5)
        W_cr = np.load(os.path.join(module_folder,'resources','cr_weights.npy'))    # shape (2048, 5)

        estimated_pupil_point = np.dot(rep, W_pupil)  # shape (5,)
        estimated_cr_point = np.dot(rep, W_cr)   # shape (5,)

        x_pupil, y_pupil = estimated_pupil_point[:2]*np.array([640,480])
        x_pupil = int(x_pupil)
        y_pupil = int(y_pupil)
        logging.debug("estimated pupil point is ({0},{1})".format(x_pupil,y_pupil))
        print "estimated pupil point is ({0},{1})".format(x_pupil,y_pupil)
        # bbox is xmin, xmax, ymin, ymax
        # x, y = 320, 240
        bbox_pupil = [x_pupil-margin, x_pupil+margin, y_pupil-margin, y_pupil+margin]

        x_cr, y_cr = estimated_cr_point[:2]
        x_cr = int(x_cr)
        y_cr = int(y_cr)
        logging.debug("estimated cr point is ({0},{1})".format(x_cr,y_cr))
        print "estimated cr point is ({0},{1})".format(x_cr,y_cr)

        # bbox is xmin, xmax, ymin, ymax
        bbox_cr = [x_cr-margin, x_cr+margin, y_cr-margin, y_cr+margin]
        # bbox_cr = None

        # plot bbox on mean_frame for QC check
        mean_frame_annotated = np.dstack([self.mean_frame, self.mean_frame, self.mean_frame])
        mean_frame_annotated = self.annotate_frame_with_bbox(mean_frame_annotated,pupil_bbox=bbox_pupil,cr_bbox=bbox_cr)
        mean_frame_annotated = self.annotate_frame_with_point(mean_frame_annotated,pupil=(x_pupil, y_pupil),cr=(x_cr, y_cr))

        dpi = 100.0
        fig, ax = plt.subplots(figsize=(mean_frame.shape[1]/dpi, mean_frame.shape[0]/dpi))
        fig.subplots_adjust(left=0,right=1,bottom=0,top=1)

        ax.imshow(mean_frame_annotated, aspect='normal')
        ax.axis('off')
        fig.savefig(os.path.join(self.qc_folder, 'mean_frame_annotated.'+image_type), dpi=dpi)

        self.bbox_pupil = bbox_pupil
        self.bbox_cr = bbox_cr

        return bbox_pupil, bbox_cr

    def compute_mean_frame(self, image_file_type='png'):
        logging.debug("computing mean frame")

        mean_frame = np.zeros(self.im_shape)

        frames_read = 0
        for input_frame in self.input_stream:
            mean_frame += input_frame
            frames_read += 1

        mean_frame /= frames_read
        mean_frame = mean_frame.astype(np.uint8)

        np.save(self.mean_frame_file, mean_frame)

        dpi = 100.0
        fig, ax = plt.subplots(figsize=(mean_frame.shape[1]/dpi, mean_frame.shape[0]/dpi))
        fig.subplots_adjust(left=0,right=1,bottom=0,top=1)
        ax.imshow(mean_frame, aspect='normal', cmap='gray')
        ax.axis('off')
        fig.savefig(os.path.join(self.qc_folder, 'mean_frame.' + image_file_type), dpi=dpi)
        plt.close()

        if self.bbox_cr and self.bbox_pupil:
            mean_frame_annotated = np.dstack([mean_frame,mean_frame,mean_frame])
            mean_frame_annotated = self.annotate_frame_with_bbox(mean_frame_annotated,pupil_bbox=self.bbox_pupil,cr_bbox=self.bbox_cr)        

            fig, ax = plt.subplots(figsize=(mean_frame.shape[1]/dpi, mean_frame.shape[0]/dpi))
            fig.subplots_adjust(left=0,right=1,bottom=0,top=1)
            ax.imshow(mean_frame_annotated, aspect='normal')
            ax.axis('off')
            fig.savefig(os.path.join(self.qc_folder, 'mean_frame_bbox.' + image_file_type), dpi=dpi)
            plt.close()

        return mean_frame



    def detect_eye_closed(self):

        try:
            import keras
        except ImportError:
            logging.debug("keras failed to import.  Can't detect eye closure")
            return None

        from keras.applications import InceptionV3

        logging.debug("Detecting eye closed frames")
        # compute the representation for mean_frame
        model = InceptionV3(include_top=False, weights='imagenet')
        # print self.mean_frame.dtype, self.mean_frame.shape

        # get pre-trained svm
        from sklearn.externals import joblib
        module_folder = os.path.dirname(os.path.abspath(__file__))
        svm = joblib.load(os.path.join(module_folder, 'resources','svm_trained.pkl'))

        def compute_rep(frame):
            mp_temp = frame.astype(np.float32)
            mp_temp -= 128
            mp_temp /= 128
            rep = model.predict(mp_temp.reshape((1,)+mp_temp.shape))  # shape (1,13,18,2048)
            rep[rep<0]=0
            rep = np.mean(rep, axis=(0,1,2))  # shape (2048,)

            return rep

        is_closed = np.zeros(self.num_frames)

        for input_frame in self.input_stream:
            rep = compute_rep(input_frame)
            is_closed[i] = svm.predict(rep.reshape(-1,len(rep)))[0]

        self.is_closed = is_closed
        save_path = os.path.join(self.results_folder, 'is_closed.npy')

        logging.debug("Saving is_closed to:")
        logging.debug("\t%s", save_path)
        #
        np.save(save_path, self.is_closed)

        return is_closed

    def process_movie(self, movie_output_stream=None,
                      output_frames=False,
                      output_annotation_frames=False,
                      image_file_type = 'jpg' ):

        # these aren't really used yet.
        # self.pupil_loc = (0,0)
        # self.cr_loc = (0,0)

        self.pupil_params = np.zeros([self.num_frames, 5])
        self.cr_params = np.zeros([self.num_frames, 5])

        if movie_output_stream:
            movie_output_stream.open(self.annotated_movie_file)
        else:
            movie_output_stream = None

        if output_frames:
            frame_output_stream = ImageOutputStream()
            frame_output_stream.open(os.path.join(self.input_image_folder, 'input_frame-%06d.'+image_file_type))
        else:
            frame_output_stream = None

        if output_annotation_frames:
            annotation_frame_output_stream = ImageOutputStream()
            annotation_frame_output_stream.open(os.path.join(self.frames_folder, 'output_frame-%06d.'+image_file_type))
        else:
            annotation_frame_output_stream = None

        for i, input_frame in enumerate(self.input_stream):
            # get pupil and corneal reflection parameters, this line is the actual eye tracking algorithm
            pupil, cr = self.process_image(input_frame, bbox_pupil=self.bbox_pupil, bbox_cr=self.bbox_cr)

            pupil_params = (pupil[0][0],pupil[0][1],pupil[1],pupil[2][0],pupil[2][1])
            cr_params = (cr[0][0],cr[0][1],cr[1],cr[2][0],cr[2][1])

            if frame_output_stream:
                frame_output_stream.write(input_frame)

            if movie_output_stream or annotation_frame_output_stream:
                annotated_frame = self.annotate_frame(np.dstack([input_frame,input_frame,input_frame]), 
                                                      pupil_params, 
                                                      cr_params)

                if movie_output_stream:
                    movie_output_stream.write( annotated_frame )

                if annotation_frame_output_stream:
                    annotation_frame_output_stream.write( annotated_frame )

            # save results in arrays
            self.pupil_params[i] = (pupil[0][0],pupil[0][1],pupil[1],pupil[2][0],pupil[2][1])
            self.cr_params[i] = (cr[0][0],cr[0][1],cr[1],cr[2][0],cr[2][1])

            if i % 100 == 0:
                logging.debug("tracked frame %d", i)

        logging.debug("Saving pupil and cr parameters to:")
        logging.debug("\t%s", self.pupil_file)
        logging.debug("\t%s", self.cr_file)
        #
        np.save(self.pupil_file, self.pupil_params)
        np.save(self.cr_file, self.cr_params)

        if movie_output_stream:
            movie_output_stream.close()

        if frame_output_stream:
            frame_output_stream.close()

        if annotation_frame_output_stream:
            annotation_frame_output_stream.close()

        # return mean_frame

    def clear_input_images(self):
        logging.debug("Deleting input image folder")
        shutil.rmtree(os.path.join(self.folder, 'input_images'))

    def process_image(self, im, bbox_pupil=None, bbox_cr=None):
        # let's try median filtering the image first
        im = medfilt2d(im, kernel_size=3)

        # find pupil and corneal reflection if auto==True
        if self.auto:
            self.pupil_loc = initial_pupil_point(im, bbox=bbox_pupil)
            self.cr_loc = initial_cr_point(im, bbox=bbox_cr)


        # find rays projecting from seed point
        pupil_rays, pupil_ray_values = generate_rays(im,self.pupil_loc)

        # save values for analysis
        self.pupil_rays = pupil_rays
        self.pupil_ray_values = pupil_ray_values

        # code for finding pupil ellipse, start with candidate points from rays
        pupil_candidate_points = self.get_candidate_points(self.pupil_rays,self.pupil_ray_values,self.threshold_factor,above_threshold=True)

        # fit pupil ellipse with all candidate points
        #pupil_params = fit_ellipse(pupil_candidate_points)

        # fit pupil ellipse with ransac algorithm
        fe=FitEllipse(10,10,0.0001,4)
        result = fe.ransac_fit(pupil_candidate_points)

        # if np.any(np.isnan(result)):    #should use np.any(np.isnan(result))
        #     pupil_params = result  #fe.ransac_fit(pupil_candidate_points)
        # else:
        #     pupil_params = ((np.nan,np.nan),np.nan,(np.nan,np.nan))  #  np.nan*np.ones(5)


        if result!=None:    #should use np.any(np.isnan(result))
            pupil_params = result  #fe.ransac_fit(pupil_candidate_points)
        else:
            logging.debug("No good fit found")
            pupil_params = ((np.nan,np.nan),np.nan,(np.nan,np.nan))  #  np.nan*np.ones(5)


        # code for finding corneal reflection, start with finding rays from center of cr
        cr_rays, cr_ray_values = generate_rays(im,self.cr_loc)

        self.cr_rays = cr_rays
        self.cr_ray_values = cr_ray_values

        cr_candidate_points = self.get_candidate_points(self.cr_rays,self.cr_ray_values,0.75,above_threshold=False)

        try:
            #cr_params = fit_ellipse(cr_candidate_points)
            fe=FitEllipse(10,10,0.0001,4)
            result = fe.ransac_fit(cr_candidate_points)

            if result!=None:
                cr_params = result #fe.ransac_fit(cr_candidate_points)
            else:
                logging.debug("No good fit found")
                cr_params = ((np.nan,np.nan),np.nan,(np.nan,np.nan))
        except Exception as e:
            logging.error("Error during fit: %s", e.message)
            cr_params = ((np.nan,np.nan),np.nan,(np.nan,np.nan))

        # update instance variables
        if not np.isnan(pupil_params[0][0]):   #should use np.any(np.isnan(result))
            self.pupil_loc = (int(pupil_params[0][1]),int(pupil_params[0][0]))

        self.pupil_candidate_points = pupil_candidate_points
        self.cr_candidate_points = cr_candidate_points

        return pupil_params, cr_params

    def get_candidate_points(self,rays,ray_values,threshold_f,above_threshold=True):

        candidate_points = []
        # find candidate points for ellipse from threshold crossing of the image over the rays
        for i, ray in enumerate(rays):

            sample_ray = ray_values[i][:self.cutoff_pixels]
            threshold = threshold_f*np.mean(sample_ray)

            for t,g in enumerate(ray_values[i][self.cutoff_pixels:]):
                if above_threshold:
                    if g > threshold:
                        new_point = ray.T[t+self.cutoff_pixels]
                        candidate_points += [new_point]
                        break
                else:
                    if g < threshold:
                        new_point = ray.T[t+self.cutoff_pixels]
                        candidate_points += [new_point]
                        break

        return candidate_points

    def set_seed_points(self, initial_pupil_x, initial_pupil_y, initial_cr_x, initial_cr_y):
        self.initial_pupil_x = initial_pupil_x
        self.initial_pupil_y = initial_pupil_y
        self.initial_cr_x = initial_cr_x
        self.initial_cr_y = initial_cr_y

    def process_all_images(self):
        """ deprecated """

        # these aren't really used yet.
        self.pupil_loc = (0,0)
        self.cr_loc = (0,0)

        frame_list = os.listdir(self.input_image_folder)
        num_frames = len(frame_list)

        self.pupil_params = np.zeros([num_frames, 5])
        self.cr_params = np.zeros([num_frames, 5])

        for i,frame in enumerate(frame_list):
            logging.debug("Processing frame %d", i)
            if frame[-4:]!='.jpg' and frame[-4:]!='.png':  continue  # just in case some OS specific files snuck in (like in OS X)
            frame_path = os.path.join(self.input_image_folder,frame)

            # open Image, convert to gray scale and then to numpy array
            im = Image.open(frame_path)
            im = im.convert('L')
            im = np.array(im)


            # get pupil and corneal reflection parameters, this line is the actual eye tracking algorithm
            pupil, cr = self.process_image(im)

            # save results in arrays
            self.pupil_params[i] = (pupil[0][0],pupil[0][1],pupil[1],pupil[2][0],pupil[2][1])
            self.cr_params[i] = (cr[0][0],cr[0][1],cr[1],cr[2][0],cr[2][1])

        logging.debug("Saving pupil and cr parameters to:")
        logging.debug("\t%s", self.pupil_file)
        logging.debug("\t%s", self.cr_file)

        np.save(self.pupil_file, self.pupil_params)
        np.save(self.cr_file, self.cr_params)


    @staticmethod
    def rotate(X, Y, center_x, center_y, theta):

        Xp = (X-center_x)*np.cos(theta) - (Y-center_y)*np.sin(theta) + center_x
        Yp = (X-center_x)*np.sin(theta) + (Y-center_y)*np.cos(theta) + center_y

        return Xp, Yp

    @staticmethod
    def get_ellipse_mask(X, Y, params):

        center_x, center_y, theta, axis1, axis2 = params

        dX = X - center_x
        dY = Y - center_y

        theta = theta*np.pi/180.

        Xp = dX*np.cos(theta) + dY*np.sin(theta)
        Yp = -dX*np.sin(theta) + dY*np.cos(theta)

        mask1 = (Xp/axis1)**2 + (Yp/axis2)**2 < 1 + 0.1
        mask2 = (Xp/axis1)**2 + (Yp/axis2)**2 > 1 - 0.1

        mask = np.logical_and(mask1, mask2)

        return mask

    def annotate_frame_old(self, im, pupil=None, cr=None):

        y, x, c = im.shape

        X, Y = np.meshgrid(np.arange(x), np.arange(y))

        # pupil in red
        if pupil is not None:
            pupil_mask = self.get_ellipse_mask(X, Y, pupil)
            im.T[0].T[pupil_mask] = 255
            im.T[1].T[pupil_mask] = 0
            im.T[2].T[pupil_mask] = 0

        # cr in blue
        if cr is not None:
            cr_mask = self.get_ellipse_mask(X, Y, cr)
            im.T[0].T[cr_mask] = 0
            im.T[1].T[cr_mask] = 0
            im.T[2].T[cr_mask] = 255

        return im

    @classmethod
    def ellipse_points_from_params(cls, params):
        center_x, center_y, theta, axis1, axis2 = params
        theta = theta*np.pi/180. # convert to radians

        points_x = np.array([ axis1*np.cos(phi) + center_x for phi in np.linspace(0,2*np.pi, 1000)])
        points_y = np.array([ axis2*np.sin(phi) + center_y for phi in np.linspace(0,2*np.pi, 1000)])

        points_x, points_y = cls.rotate(points_x, points_y, center_x, center_y, theta)

        return points_x, points_y

    def annotate_frame(self, im, pupil=None, cr=None):
        
        y, x, c = im.shape

        im_pil = Image.fromarray(im)

        draw = ImageDraw.Draw(im_pil)

        if pupil is not None:
            points_x, points_y = self.ellipse_points_from_params(pupil)
            draw.point(zip(points_x, points_y), fill=(255,0,0))
            # for i, px in enumerate(points_x):
            #     im[int(points_y[i]), int(px), 0] = 255

        if cr is not None:
            points_x, points_y = self.ellipse_points_from_params(cr)
            draw.point(zip(points_x, points_y), fill=(0,0,255))
            # for i, px in enumerate(points_x):
            #     im[int(points_y[i]), int(px), 0] = 255

        # return im
        return np.array(im_pil)

    def annotate_frame_with_bbox(self, im, pupil_bbox=None, cr_bbox=None):

        # y, x, c = im.shape

        im_pil = Image.fromarray(im)

        draw = ImageDraw.Draw(im_pil)

        if pupil_bbox is not None:
            xmin, xmax, ymin, ymax = pupil_bbox
            # print pupil_bbox
            draw.rectangle([xmin,ymin,xmax,ymax],outline=(255,0,0))

            # for i, px in enumerate(points_x):
            #     im[int(points_y[i]), int(px), 0] = 255

        if cr_bbox is not None:
            # print cr_bbox
            xmin, xmax, ymin, ymax = cr_bbox
            draw.rectangle([xmin,ymin,xmax,ymax],outline=(0,0,255))
            # for i, px in enumerate(points_x):
            #     im[int(points_y[i]), int(px), 0] = 255

        # return im
        return np.array(im_pil)

    def annotate_frame_with_point(self, im, pupil=None, cr=None):

        # y, x, c = im.shape

        # print im.shape, im.dtype

        im_pil = Image.fromarray(im)

        draw = ImageDraw.Draw(im_pil)

        if pupil is not None:
            # points_x, points_y = ellipse_points_from_params(pupil)
            # draw.point(pupil, fill=(255,0,0))
            draw.ellipse([pupil[0]-5,pupil[1]-5,pupil[0]+5,pupil[1]+5],fill=(255,0,0))
            # for i, px in enumerate(points_x):
            #     im[int(points_y[i]), int(px), 0] = 255

        if cr is not None:
            # points_x, points_y = ellipse_points_from_params(cr)
            # draw.point(cr, fill=(0,0,255))
            draw.ellipse([cr[0]-5,cr[1]-5,cr[0]+5,cr[1]+5],fill=(0,0,255))
            # for i, px in enumerate(points_x):
            #     im[int(points_y[i]), int(px), 0] = 255

        # return im
        return np.array(im_pil)

    @staticmethod
    def get_frame_index(frame_name):

        return int(frame_name[12:-4])-1  # change 7 to 12

    # def annotate_frame(self, frame, im, pupil, cr):
    #     # this function is not done yet
    #     frame_index = self.get_frame_index(frame)
    #
    #     new_im = im.copy()
    #     pupil = self.pupil_params[frame_index]
    #     cr = self.cr_params[frame_index]
    #     new_im = self.annotate_frame(new_im, pupil, cr)
    #
    #     im_fig.set_data(new_im)
    #
    #     fig.savefig(os.path.join(self.frames_folder, input_frame), dpi=100)

    def output_annotation(self, frames_to_output=None):
        """generate a the series of images with eyetracking results superimposed"""

        fig, ax = plt.subplots(figsize=(4,3))  #, frameon=False)
        fig.subplots_adjust(left=0,right=1,bottom=0,top=1)

        ax.axis('off')
        # ax.axis('tight')

        self.pupil_params = np.load(self.pupil_file)
        self.cr_params = np.load(self.cr_file)

        if frames_to_output is None:
            frames_to_output = os.listdir(self.input_image_folder)

        first_frame = frames_to_output[0]

        frame_index = self.get_frame_index(first_frame)
        im = Image.open(os.path.join(self.input_image_folder, first_frame))
        im = np.array(im)

        new_im = np.dstack([im,im,im])
        pupil = self.pupil_params[frame_index]
        cr = self.cr_params[frame_index]
        new_im = self.annotate_frame(new_im, pupil, cr)

        im_fig = ax.imshow(new_im, aspect='normal')  #extent=(0,1,1,0)

        fig.savefig(os.path.join(self.frames_folder, first_frame), dpi=100)

        for input_frame in frames_to_output[1:]:

            frame_index = self.get_frame_index(input_frame)
            im = Image.open(os.path.join(self.input_image_folder, input_frame))
            im = np.array(im)

            new_im = np.dstack([im,im,im])
            pupil = self.pupil_params[frame_index]
            cr = self.cr_params[frame_index]
            new_im = self.annotate_frame(new_im, pupil, cr)

            im_fig.set_data(new_im)

            fig.savefig(os.path.join(self.frames_folder, input_frame), dpi=100)

    def output_QC(self, image_type='png'):
        """generate a set of summary statistics and plots for QC purposes"""

        logging.debug("saving QC images")
        self.pupil_params = np.load(self.pupil_file)
        self.cr_params = np.load(self.cr_file)

        logging.debug("saving pupil position")
        fig, ax = plt.subplots(1)
        ax.plot(self.pupil_params.T[0], label='pupil x')
        ax.plot(self.pupil_params.T[1], label='pupil y')
        ax.set_xlabel('frame index')
        ax.set_title('pupil position')
        ax.legend()
        fig.savefig(os.path.join(self.qc_folder, 'pupil_position.'+image_type))

        logging.debug("saving cr position")
        fig, ax = plt.subplots(1)
        ax.plot(self.cr_params.T[0], label='cr x')
        ax.plot(self.cr_params.T[1], label='cr y')
        ax.set_xlabel('frame index')
        ax.set_title('CR position')
        ax.legend()
        fig.savefig(os.path.join(self.qc_folder, 'cr_position.'+image_type))

        logging.debug("saving pupil axes")
        fig, ax = plt.subplots(1)
        ax.plot(self.pupil_params.T[3], label='pupil axis 1')
        ax.plot(self.pupil_params.T[4], label='pupil axis 2')
        ax.set_xlabel('frame index')
        ax.set_title('Pupil major and minor axis size')
        ax.legend()
        fig.savefig(os.path.join(self.qc_folder, 'pupil_axes.'+image_type))

        logging.debug("saving cr major/minor axis")
        fig, ax = plt.subplots(1)
        ax.plot(self.cr_params.T[3], label='cr axis 1')
        ax.plot(self.cr_params.T[4], label='cr axis 2')
        ax.set_xlabel('frame index')
        ax.set_title('CR major and minor axis size')
        ax.legend()
        fig.savefig(os.path.join(self.qc_folder, 'cr_axes.'+image_type))

        logging.debug("saving pupil angle")
        fig, ax = plt.subplots(1)
        ax.plot(self.pupil_params.T[2], label='pupil angle')
        ax.set_xlabel('frame index')
        ax.set_title('pupil major axis angle')
        ax.legend()
        fig.savefig(os.path.join(self.qc_folder, 'pupil_angle.'+image_type))

        logging.debug("saving cr angle")
        fig, ax = plt.subplots(1)
        ax.plot(self.cr_params.T[2], label='cr angle')
        ax.set_xlabel('frame index')
        ax.set_title('corneal reflection major axis angle')
        ax.legend()
        fig.savefig(os.path.join(self.qc_folder, 'cr_angle.'+image_type))


        logging.debug("computing density")
        # the remainder of these take a *very* long time
        T = self.pupil_params.shape[0]
        y, x = self.im_shape

        mean_frame = np.dstack([self.mean_frame,self.mean_frame,self.mean_frame])
        pupil_density = np.zeros((y, x, 3))
        # pupil_all = 255*np.ones(pupil_density.shape, np.uint8)
        # pupil_all = np.stack([mean_frame, mean_frame, mean_frame], axis=2)
        pupil_all = mean_frame.copy()

        cr_density = np.zeros((y, x, 3))
        # cr_all = 255*np.ones(cr_density.shape, np.uint8)
        # cr_all = np.stack([mean_frame, mean_frame, mean_frame], axis=2)
        cr_all = mean_frame.copy()
        temp = np.zeros((y,x,3), dtype=np.uint8)
        for t in range(T):
            if t % 100 == 0:
                logging.debug("finished %d frames", t)
            ptemp = self.annotate_frame(temp.copy(), self.pupil_params[t])
            pupil_density += ptemp #, self.cr_params[t])
            pupil_all = self.annotate_frame(pupil_all, self.pupil_params[t])

            crtemp = self.annotate_frame(temp.copy(), cr=self.cr_params[t])
            cr_density += crtemp #, self.cr_params[t])
            cr_all = self.annotate_frame(cr_all, cr=self.cr_params[t])

        logging.debug("plotting pupil density")
        fig, ax = plt.subplots(1)
        ax.imshow(np.log(1+pupil_density[:,:,0]), cmap='Greys', interpolation='nearest')
        # ax.axis('off')
        ax.set_title('Pupil ellipse density')
        fig.savefig(os.path.join(self.qc_folder, 'pupil_density.'+image_type))


        logging.debug("plotting pupil all")
        fig, ax = plt.subplots(1)
        ax.imshow(pupil_all, cmap='Greys', interpolation='nearest')
        # ax.axis('off')
        ax.set_title('All pupil ellipses combined')
        fig.savefig(os.path.join(self.qc_folder, 'pupil_all_plot.'+image_type))

        logging.debug("plotting cr density")
        fig, ax = plt.subplots(1)
        ax.imshow(np.log(1+cr_density[:,:,2]), cmap='Greys', interpolation='nearest')
        # ax.axis('off')
        ax.set_title('CR ellipse density')
        fig.savefig(os.path.join(self.qc_folder, 'cr_density.'+image_type))

        logging.debug("plotting cr all")
        fig, ax = plt.subplots(1)
        ax.imshow(cr_all, cmap='Greys', interpolation='nearest')
        ax.set_title('All CR ellipses combined')
        # ax.axis('off')
        fig.savefig(os.path.join(self.qc_folder, 'cr_all_plot.'+image_type))




import sys
from subprocess import call
from pkg_resources import resource_filename #@UnresolvedImport

the_script = resource_filename(__name__, 'run_optimize.sh')

cmd = ['/bin/bash', the_script]
cmd.extend(sys.argv[1:])

print(' '.join(cmd))

call(cmd)


import os
import sys
from allensdk.internal.api.queries.biophysical_module_api \
    import BiophysicalModuleApi
from allensdk.internal.api.queries.biophysical_module_reader \
    import BiophysicalModuleReader

nmr_id = sys.argv[-1]
bma = BiophysicalModuleApi('http://lims2')
result = bma.get_neuronal_model_runs(nmr_id)
r = BiophysicalModuleReader()
r.read_lims_message(result, None)

def well_known_file_path(d):
    return os.path.join(d['storage_directory'], d['filename'])

print(os.path.join(r.neuronal_model_run_dir(),
      'EPHYS_BIOPHYS_SIMULATE_QUEUE_' + nmr_id + '_input.json'))
print(r.stimulus_path())
print(r.fit_parameters_path())
print(r.morphology_path())
print('\n'.join(r.mod_file_paths()))




"""Module for calculating annotated region metrics from ISI data"""
import numpy as np
import logging

# These scaling factors are derived from experimental geometry using
# screen width 52.0 cm, screen height 32.0 cm (16:10 24 inch monitor)
# mouse 10.0 cm from center of monitor
ALTITUDE_SCALE = 0.322
AZIMUTH_SCALE = 0.383


def eccentricity(az, alt, az_center, alt_center):
    """Compute eccentricity
    
    Parameters
    ----------
    az : numpy.ndarray
        Azimuth retinotopic map
    alt : numpy.ndarray
        Altitude retinotopic map
    az_center : float
        Azimuth value to use as center of eccentricity map
    alt_center : float
        Altitude value to use as center of eccentricity map

    Returns
    -------
    numpy.ndarray
        Eccentricity map
    """
    daz = az - az_center
    dalt = alt - alt_center
    ecc = np.arctan(np.sqrt(np.square(np.tan(dalt)) +
                            np.square(np.tan(daz))/np.square(np.cos(dalt))))
    return ecc


def retinotopy_metric(mask, isi_map):
    """Compute retinotopic metrics for a responding area
    
    Parameters
    ----------
    mask : numpy.ndarray
        Mask representing the area over which to calculate metrics
    isi_map : numpy.ndarray
        Retinotopic map

    Returns
    -------
    (float, float, float, float) tuple
        min, max, range, bias of retinotopic map over masked region
    """
    ind = np.where( mask > 0 )
    vals = isi_map[ind]
    maxv = np.degrees(np.max(vals))
    minv = np.degrees(np.min(vals))
    ret_range = float(maxv - minv)
    ret_bias = float(abs(minv + maxv))
    return float(minv), float(maxv), ret_range, ret_bias


def create_region_mask(image_shape, x, y, width, height, mask):
    """Create mask for region on retinotopic map

    Parameters
    ----------
    image_shape : tuple
        (height, width) of retinotopic map
    x : int
        x offset of region mask within retinotopic map
    y : int
        y offset of region mask within retinotopic map
    width : int
        width of region mask
    height : int
        height of region mask
    mask : list
        region mask as a list of lists
    
    Returns
    -------
    numpy.ndarray
        Region mask
    """
    bb = np.zeros((height,width), dtype=np.uint8)
    bb[np.asarray(mask)] = 1
    region_mask = np.zeros(image_shape, dtype=np.uint8)
    region_mask[y:y + height,x:x + width] = bb
    return region_mask


def get_metrics(altitude_phase, azimuth_phase, x=None, y=None, width=None,
                height=None, mask=None, altitude_scale=ALTITUDE_SCALE,
                azimuth_scale=AZIMUTH_SCALE):
    """Calculate annotated region metrics"""
    altitude = altitude_phase * altitude_scale
    azimuth = azimuth_phase * azimuth_scale

    eccentricity_ret_zero = np.degrees(
        eccentricity(azimuth, altitude, 0.0, 0.0))

    result = {}

    region_mask = create_region_mask(altitude.shape, x, y, width, height,
                                     mask)

    # compute centroid
    centroid = [np.mean(x_value) for x_value in np.where(region_mask)]
    result['y_centroid'] = centroid[0]
    result['x_centroid'] = centroid[1]

    # compute azimuth/altitude max,min,range and bias
    az_min, az_max, az_range, az_bias = retinotopy_metric(region_mask,
                                                          azimuth)
    alt_min, alt_max, alt_range, alt_bias = retinotopy_metric(region_mask,
                                                              altitude)
    result['azimuth_min'] = az_min
    result['azimuth_max'] = az_max
    result['azimuth_range'] = az_range
    result['azimuth_bias'] = az_bias
    result['altitude_min'] = alt_min
    result['altitude_max'] = alt_max
    result['altitude_range'] = alt_range
    result['altitude_bias'] = alt_bias

    # eccentricity at centroid
    cy = round(result['y_centroid'])
    cx = round(result['x_centroid'])
    result['eccentricity_at_centroid'] = float(eccentricity_ret_zero[cy,cx])

    return result



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

from allensdk.api.queries.rma_template import RmaTemplate

class BiophysicalModuleApi(RmaTemplate):
    ''''''
    
    rma_templates = \
        {"biophysical_lims_queries": [
            {'name': 'neuronal_model_runs_by_ids',
             'description': 'see name',
             'model': 'NeuronalModelRun',
             'criteria': '[id$in{{ neuronal_model_run_ids }}]',
             'include': 'well_known_files(well_known_file_type),'
                        'neuronal_model(well_known_files(well_known_file_type),'
                            'specimen(project,specimen_tags,'
                                'ephys_roi_result'
                                '(ephys_qc_criteria,'
                                'well_known_files(well_known_file_type)),'
                                'neuron_reconstructions'
                                '(well_known_files(well_known_file_type)),'
                                'ephys_sweeps'
                                '(ephys_sweep_tags,'
                                'ephys_stimulus(ephys_stimulus_type))),'
                                'neuronal_model_template'
                                '(neuronal_model_template_type,'
                                'well_known_files(well_known_file_type)))',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['neuronal_model_run_ids']
            },
            {'name': 'neuronal_models_by_ids',
             'description': 'see name',
             'model': 'NeuronalModel',
             'criteria': '[id$in{{ neuronal_model_ids }}]',
             'include': 'well_known_files(well_known_file_type),'
                        'specimen(project,specimen_tags,'
                            'ephys_roi_result'
                            '(ephys_qc_criteria,'
                            'well_known_files(well_known_file_type)),'
                            'neuron_reconstructions'
                            '(well_known_files(well_known_file_type)),'
                            'ephys_sweeps'
                            '(ephys_sweep_tags,'
                            'ephys_stimulus(ephys_stimulus_type))),'
                            'neuronal_model_template'
                            '(neuronal_model_template_type,'
                            'well_known_files(well_known_file_type))',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['neuronal_model_ids']
            }                                      
        ]}
    
    
    def __init__(self, base_uri=None):
        super(BiophysicalModuleApi, self).__init__(base_uri,
                                                   query_manifest=BiophysicalModuleApi.rma_templates)
    
    
    def get_neuronal_model_runs(self, neuronal_model_run_ids=None):
        '''List Neuronal Model Rusn available through LIMS
        with associated info needed to run in NEURON.
        
        Parameters
        ----------
        neuronal_model_run_ids : integer or list of integers, optional
            only select specific neuronal_model_runs.
        
        Returns
        -------
        dict : neuronal model run metadata        
        '''
        data = self.template_query('biophysical_lims_queries',
                                   'neuronal_model_runs_by_ids',
                                   neuronal_model_run_ids=neuronal_model_run_ids)
        
        return data
    
    
    def get_neuronal_models(self, neuronal_model_ids=None):
        '''List Neuronal Models available through LIMS
        with associated info needed to run in NEURON.
        
        Parameters
        ----------
        neuronal_model_ids : integer or list of integers, optional
            only select specific neuronal_models.
        
        Returns
        -------
        dict : neuronal model metadata        
        '''
        data = self.template_query('biophysical_lims_queries',
                                   'neuronal_models_by_ids',
                                   neuronal_model_ids=neuronal_model_ids)
        
        return data    


import os
import sys
import subprocess
import numpy as np
import allensdk.internal.model.biophysical.ephys_utils as ephys_utils
import passive_fitting.preprocess as passive_prep
import allensdk.core.json_utilities as ju
from allensdk.model.biophys_sim.config import Config
from allensdk.core.nwb_data_set import NwbDataSet
from allensdk.internal.model.biophysical.passive_fitting import neuron_passive_fit
from allensdk.internal.model.biophysical.passive_fitting import neuron_passive_fit2
from allensdk.internal.model.biophysical.passive_fitting import neuron_passive_fit_elec
from pkg_resources import resource_filename #@UnresolvedImport
import logging
import logging.config as lc


_run_passive_fit_log = logging.getLogger('allensdk.internal.model.biophysical.run_passive_fit')


def run_passive_fit(description):
    output_directory = description.manifest.get_path('WORKDIR')
    neuronal_model = ju.read(description.manifest.get_path('neuronal_model_data'))
    specimen_data = neuronal_model['specimen']
    
    is_spiny = not any(t['name'] == u'dendrite type - aspiny' for t in specimen_data['specimen_tags'])
    
    all_sweeps = specimen_data['ephys_sweeps']
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    cap_check_sweeps, _, _ = \
        ephys_utils.get_sweeps_of_type('C1SQCAPCHK',
                                       all_sweeps)
    
    passive_fit_data = {}

    if len(cap_check_sweeps) > 0:
        data_set = NwbDataSet(description.manifest.get_path('stimulus_path'))
        d = passive_prep.get_passive_fit_data(cap_check_sweeps, data_set);

        grand_up_file = os.path.join(output_directory, 'upbase.dat')
        np.savetxt(grand_up_file, d['grand_up'])

        grand_down_file = os.path.join(output_directory, 'downbase.dat')
        np.savetxt(grand_down_file, d['grand_down'])
        
        passive_fit_data["bridge"] = d['bridge_avg']
        passive_fit_data["escape_time"] = d['escape_t']

        fit_1_file = description.manifest.get_path('fit_1_file')
        fit_1_params = subprocess.check_output([sys.executable,
                                                '-m', neuron_passive_fit.__name__, 
                                                str(d['escape_t']),
                                                os.path.realpath(description.manifest.get_path('manifest')) ])
        passive_fit_data['fit_1'] = ju.read(fit_1_file)

        fit_2_file = description.manifest.get_path('fit_2_file')

        fit_2_params = subprocess.check_output([sys.executable,
                                                '-m', neuron_passive_fit2.__name__,
                                                str(d['escape_t']),
                                                os.path.realpath(description.manifest.get_path('manifest')) ])
        passive_fit_data['fit_2'] = ju.read(fit_2_file)

        fit_3_file = description.manifest.get_path('fit_3_file')
        fit_3_params = subprocess.check_output([sys.executable,
                                                '-m', neuron_passive_fit_elec.__name__,
                                                str(d['escape_t']),
                                                str(d['bridge_avg']),
                                                str(1.0),
                                                os.path.realpath(description.manifest.get_path('manifest')) ])
        passive_fit_data['fit_3'] = ju.read(fit_3_file)
        
        # Check for potentially problematic outcomes
        cm_rel_delta = (passive_fit_data["fit_1"]["Cm"] - passive_fit_data["fit_3"]["Cm"]) / passive_fit_data["fit_1"]["Cm"]
        if passive_fit_data["fit_2"]["err"] < passive_fit_data["fit_1"]["err"]:
            _run_passive_fit_log.debug("Fixed Ri gave better results than original")
            if passive_fit_data["fit_2"]["err"] < passive_fit_data["fit_3"]["err"]:
                _run_passive_fit_log.debug("Using fixed Ri results")
                passive_fit_data["fit_for_next_step"] = passive_fit_data["fit_2"]
            else:
                _run_passive_fit_log.debug("Using electrode results")
                passive_fit_data["fit_for_next_step"] = passive_fit_data["fit_3"]
        elif abs(cm_rel_delta) > 0.1:
            _run_passive_fit_log.debug("Original and electrode fits not in sync:")
            _run_passive_fit_log.debug("original Cm: " + str(passive_fit_data["fit_1"]["Cm"]))
            _run_passive_fit_log.debug("w/ electrode Cm: " + str(passive_fit_data["fit_3"]["Cm"]))
            if passive_fit_data["fit_1"]["err"] < passive_fit_data["fit_3"]["err"]:
                _run_passive_fit_log.debug("Original has lower error")
                passive_fit_data["fit_for_next_step"] = passive_fit_data["fit_1"]
            else:
                _run_passive_fit_log.debug("Electrode has lower error")
                passive_fit_data["fit_for_next_step"] = passive_fit_data["fit_3"]
        else:
            passive_fit_data["fit_for_next_step"] = passive_fit_data["fit_1"]

        ra = passive_fit_data["fit_for_next_step"]["Ri"]
        if is_spiny:
            combo_cm = passive_fit_data["fit_for_next_step"]["Cm"]
            a1 = passive_fit_data["fit_for_next_step"]["A1"]
            a2 = passive_fit_data["fit_for_next_step"]["A2"]
            cm1 = 1.0
            cm2 = (combo_cm * (a1 + a2) - a1) / a2
        else:
            cm1 = passive_fit_data["fit_for_next_step"]["Cm"]
            cm2 = passive_fit_data["fit_for_next_step"]["Cm"]
    else:
        _run_passive_fit_log.debug("No cap check trace found")
        ra = 100.0
        cm1 = 1.0
        if is_spiny:
            cm2 = 2.0
        else:
            cm2 = 1.0

    passive_fit_data['ra'] = ra
    passive_fit_data['cm1'] = cm1
    passive_fit_data['cm2'] = cm2
    
    return passive_fit_data


def main(limit, manifest_path):
    app_config = Config()
    description = app_config.load(manifest_path)

    if 'LOG_CFG' in os.environ:
        log_config = os.environ['LOG_CFG']
    else:
        log_config = resource_filename('allensdk.model.biophysical',
                                       'logging.conf')
        os.environ['LOG_CFG'] = log_config
    lc.fileConfig(log_config)

    run_passive_fit(description)


if __name__ == "__main__":
    limit = sys.argv[-2]
    manifest_path = sys.argv[-1]
    
    main(limit, manifest_path)
    



#!/usr/bin/python
import os
import h5py
import sys
import shutil
import traceback
import subprocess

import nwb
from allensdk.internal.core.lims_pipeline_module import PipelineModule

# development/debugging code
#infile = "Ndnf-IRES2-dgCre_Ai14-256189.05.01-compressed.nwb"
#outfile = "foo.nwb"
#if len(sys.argv) == 1:
#    sys.argv.append(infile)
#    sys.argv.append(outfile)

# this script is meant to clone the core functionality of the 
#   existing (Igor) Hdf5->Nwb converter.
# the previous converter performed two distinct tasks. In this iteration,
#   those tasks will be split into separate modules. This module will
#   perform the file conversion. A second module will analyze the file
#   and extract sweep data

# window for leading test pulse, in seconds
PULSE_LEN = 0.1
EXPERIMENT_START_TIME = 0.75


def main():
    module = PipelineModule()
    jin = module.input_data()

    infile = jin["input_nwb"]
    outfile = jin["output_nwb"]

    # a temporary nwb file must be created. this is that file's name
    tmpfile = outfile + ".tmp"

    # create temp file and make modifications to it using h5py
    shutil.copy2(infile, tmpfile)
    f = h5py.File(tmpfile, "a")
    # change dataset names in acquisition time series to match that
    #   of existing ephys NWB files
    # also rescale the contents of 'data' fields to match the scaling
    #   in original files
    acq = f["acquisition/timeseries"]
    sweep_nums = []
    for k, v in acq.iteritems():
        # parse out sweep number
        try:
            num = int(k[5:10])
        except:
            print("Error - unexpected sweep name encountered in IGOR nwb file")
            print("Sweep called: '%s'" % k)
            print("Expecting 5-digit sweep number between chars 5 and 9")
            sys.exit(1)
        swp = "Sweep_%d" % num
        # rename objects
        try:
            acq.move(k, swp)
            ts = acq[swp]
            ts.move("stimulus_description", "aibs_stimulus_description")
        except:
            print("*** Error renaming HDF5 object in %s" % swp)
            type_, value_, traceback_ = sys.exc_info()
            print traceback.print_tb(traceback_)
            sys.exit(1)
        # rescale contents of data so conversion is 1.0
        try:
            data = ts["data"]
            scale = float(data.attrs["conversion"])
            data[...] = data.value * scale
            data.attrs["conversion"] = 1.0
        except:
            print("*** Error rescaling data in %s" % swp)
            type_, value_, traceback_ = sys.exc_info()
            print traceback.print_tb(traceback_)
            sys.exit(1)
        # keep track of sweep numbers
        sweep_nums.append("%d"%num)
        
    ###################################
    #... ditto for stimulus time series
    stim = f["stimulus/presentation"]
    for k, v in stim.iteritems():
        # parse out sweep number
        try:
            num = int(k[5:10])
        except:
            print("Error - unexpected sweep name encountered in IGOR nwb file")
            print("Sweep called: '%s'" % k)
            print("Expecting 5-digit sweep number between chars 5 and 9")
            sys.exit(1)
        swp = "Sweep_%d" % num
        try:
            stim.move(k, swp)
        except:
            print("Error renaming HDF5 group from %s to %s" % (k, swp))
            sys.exit(1)
        # rescale contents of data so conversion is 1.0
        try:
            ts = stim[swp]
            data = ts["data"]
            scale = float(data.attrs["conversion"])
            data[...] = data.value * scale
            data.attrs["conversion"] = 1.0
        except:
            print("*** Error rescaling data in %s" % swp)
            type_, value_, traceback_ = sys.exc_info()
            print traceback.print_tb(traceback_)
            sys.exit(1)
        
    f.close()

    ####################################################################
    # re-open file w/ nwb library and add indexing (epochs)
    nd = nwb.NWB(filename=tmpfile, modify=True)
    for num in sweep_nums:
        ts = nd.file_pointer["acquisition/timeseries/Sweep_" + num]
        # sweep epoch
        t0 = ts["starting_time"].value
        rate = float(ts["starting_time"].attrs["rate"])
        n = float(ts["num_samples"].value)
        t1 = t0 + (n-1) * rate
        ep = nd.create_epoch("Sweep_" + num, t0, t1)
        ep.add_timeseries("stimulus", "stimulus/presentation/Sweep_"+num)
        ep.add_timeseries("response", "acquisition/timeseries/Sweep_"+num)
        ep.finalize()
        if "CurrentClampSeries" in ts.attrs["ancestry"]:
            # test pulse epoch
            t0 = ts["starting_time"].value
            t1 = t0 + PULSE_LEN
            ep = nd.create_epoch("TestPulse_" + num, t0, t1)
            ep.add_timeseries("stimulus", "stimulus/presentation/Sweep_"+num)
            ep.add_timeseries("response", "acquisition/timeseries/Sweep_"+num)
            ep.finalize()
            # experiment epoch
            t0 = ts["starting_time"].value
            t1 = t0 + (n-1) * rate
            t0 += EXPERIMENT_START_TIME
            ep = nd.create_epoch("Experiment_" + num, t0, t1)
            ep.add_timeseries("stimulus", "stimulus/presentation/Sweep_"+num)
            ep.add_timeseries("response", "acquisition/timeseries/Sweep_"+num)
            ep.finalize()
    nd.close()

    # rescaling the contents of the data arrays causes the file to grow
    # execute hdf5-repack to get it back to its original size
    try:
        print("Repacking hdf5 file with compression")
        process = subprocess.Popen(["h5repack", "-f", "GZIP=4", tmpfile, outfile], stdout=subprocess.PIPE)
        process.wait()
    except:
        print("Unable to run h5repack on temporary nwb file")
        print("--------------------------------------------")
        raise

    try:
        print("Removing temporary file")
        os.remove(tmpfile)
    except:
        print("Unable to delete temporary file ('%s')" % tmpfile)
        raise

    # done (nothing to return)
    module.write_output_data({})



if __name__=='__main__': main()




#!/usr/bin/python
import logging
import sys
import math
import os
import re
import copy
import json
import numpy as np
import argparse
import h5py

from allensdk.internal.core.lims_pipeline_module import PipelineModule
from allensdk.core.nwb_data_set import NwbDataSet

    
def main(jin):
    # load QC criteria and sweep table from input json file
    try:
        qc_criteria = jin['ephys_qc_criteria']
        experiment_data = jin['experiment_data']
        sweep_data = jin['sweep_data']
        nwb_file = jin["nwb_file"]
    except:
        raise IOError("Input json file is missing requisite data")

    jout = {}

    # PBS-333
    # C1NSSEED stimuli have many instances, but these instances aren't
    #   stored with the sweep. Ie, the sweep stores the stimulus value
    #   C1NSSEED while the stimulus table stores C1NSSEED_2150112.
    # To address this, check the stimulus table for any instance of
    #   C1NSSEED, and if it exists, append a plain-jane "C1NSSEED" stimulus
    #   so later checks work
    for name in jin["current_clamp_stimuli"]:
        if name.startswith("C1NSSEED_"):
            jin["current_clamp_stimuli"].append("C1NSSEED")

    # list of reasons to fail entire cell. if anything is added to this list,
    #   the cell will be tagged for failure (ie, this list also serves
    #   as a 'fail' flag)
    exp_fail_tags = []

    experiment_state = {}
    jout["experiment_data"] = experiment_state

    # blowout voltage
    experiment_state["failed_blowout"] = False
    try:
        blowout = experiment_data["blowout_mv"]
        low = qc_criteria["blowout_mv_min"]
        high = qc_criteria["blowout_mv_max"]
        if blowout is None or math.isnan(blowout):
            exp_fail_tags.append("Missing blowout value (%s)" % str(blowout))
            experiment_state["failed_blowout"] = True
        if blowout < low or blowout > high:
            exp_fail_tags.append("blowout outside of range")
            experiment_state["failed_blowout"] = True
    except Exception, e:
        exp_fail_tags.append("Error analyzing blowout. " + e.message)
        experiment_state["failed_blowout"] = True


    # "electrode 0"
    experiment_state["failed_electrode_0"] = False
    try:
        e0 = experiment_data["electrode_0_pa"]
        if e0 is None or math.isnan(e0):
            exp_fail_tags.append("e0 -- missing value (%s)" % str(e0))
            experiment_state["failed_electrode_0"] = True
        if abs(e0) > qc_criteria["electrode_0_pa_max"]:
            exp_fail_tags.append("e0 -- exceeds max")
            experiment_state["failed_electrode_0"] = True
    except Exception, e:
        exp_fail_tags.append("Error analyzing blowout. " + e.message)
        experiment_state["failed_electrode_0"] = True


    # measure clamp seal
    experiment_state["failed_seal"] = False
    try:
        seal = experiment_data["seal_gohm"]
        if seal is None or math.isnan(seal):
            exp_fail_tags.append("Invalid seal (%s)" % str(seal))
            experiment_state["failed_seal"] = True
        if seal < qc_criteria["seal_gohm_min"]:
            tgt = qc_criteria["seal_gohm_min"]
            reason = "%f versus criteria=%f" % (seal, tgt)
            exp_fail_tags.append("Seal (%s)" % reason)
            experiment_state["failed_seal"] = True
    except Exception, e:
        seal = None
        msg = "Seal 0 is not available. %s" % e.message
        logging.warning(msg)
        exp_fail_tags.append(msg)
        experiment_state["failed_seal"] = True


    # input and access resistance
    sr_tags = []

    try:
        sir_ratio = experiment_data['input_access_resistance_ratio']
        #r = experiment_data['input_resistance_mohm']
    except:
        sr_tags.append("Resistance ratio not available")

    try:
        sr = experiment_data['initial_access_resistance_mohm']
    except:
        sr_tags.append("Initial access resistance not available")

    try:
        if len(sr_tags) == 0:
            experiment_state["failed_bad_rs"] = False

            if sr < qc_criteria["access_resistance_mohm_min"]:
                experiment_state["failed_bad_rs"] = True
                tgt = qc_criteria["access_resistance_mohm_min"]
                reason = "%f versus criteria=%f" % (sr, tgt)
                sr_tags.append("access-resistance low (%s)" % reason)
            elif sr > qc_criteria["access_resistance_mohm_max"]:
                experiment_state["failed_bad_rs"] = True
                tgt = qc_criteria["access_resistance_mohm_max"]
                reason = "%f versus criteria=%f" % (sr, tgt)
                sr_tags.append("access-resistance high (%s)" % reason)

            if sir_ratio > qc_criteria["input_vs_access_resistance_min"]:
                experiment_state["failed_bad_rs"] = True
                tgt = qc_criteria["input_vs_access_resistance_min"]
                reason = "%f versus criteria=%f" % (sir_ratio, tgt)
                sr_tags.append("input/access resistance (%s)" % reason)
    except Exception, e:
        exp_fail_tags.append("Error analyzing access resistance. " + e.message)

    if len(sr_tags) > 0:
        exp_fail_tags.extend(sr_tags)


    experiment_state["fail_tags"] = exp_fail_tags


    ####################################################################
    # check features for each sweep
    sweep_state = {}
    jout["sweep_state"] = sweep_state
    for name, sweep in jin["sweep_data"].iteritems():
        try:
            # keep track of failures
            fail_tags = []

            sweep_num = sweep["sweep_number"]

            stim = sweep["ephys_stimulus"]["description"]
            if stim.endswith("_DA_0"):
                stim = stim[:-5]
            unit = sweep["stimulus_units"]
            # determine if sweep is current or voltage clamp
            # name may end in "[#]", so strip out section after open bracket
            stim_short = stim.split('[')[0]
            if stim_short in jin["voltage_clamp_stimuli"]:
                if unit != "Volts" and unit != "mV":
                    msg = "%s (%s) in wrong mode -- expected voltage clamp" % (name, stim)
                    fail_tags.append(msg)
            elif stim_short in jin["current_clamp_stimuli"]:
                if unit != "Amps" and unit != "pA":
                    msg = "%s (%s) in wrong mode -- expected current clamp" % (name, stim)
                    fail_tags.append(msg)
            else:
                fail_tags.append("%s has unrecognized stimulus (%s)" % (name, stim))

            if unit == "Volts" or unit == "mV":
                continue    # no QC on voltage clamp

            if len(fail_tags) > 0:
                sweep_state[name] = {}
                sweep_state[name]["state"] = "Fail"
                sweep_state[name]["reasons"] = fail_tags
                continue
            
            # pull data streams from file (this is for detecting truncated
            #   sweeps)
            sweep_data = NwbDataSet(nwb_file).get_sweep(sweep_num)
            volts = sweep_data['response']
            current = sweep_data['stimulus']
            hz = sweep_data['sampling_rate']
            idx_start, idx_stop = sweep_data['index_range']

            if sweep["pre_noise_rms_mv"] > qc_criteria["pre_noise_rms_mv_max"]:
                fail_tags.append("pre-noise")

            # check Vm and noise at end of recording
            # only do so if acquisition not truncated 
            # do not check for ramps, because they do not have 
            #   enough time to recover
            is_ramp = stim.startswith('C1RP')
            if is_ramp:
                logging.info("sweep %d skipping vrest criteria on ramp", sweep_num)
            else:
                # measure post-stimulus noise
                sweep_not_truncated = ( idx_stop == len(current) - 1 )
                if sweep_not_truncated:
                    post_noise_rms_mv = sweep["post_noise_rms_mv"]
                    if post_noise_rms_mv > qc_criteria["post_noise_rms_mv_max"]:
                        fail_tags.append("post-noise")
                else:
                    fail_tags.append("Truncated sweep")

            if sweep["slow_noise_rms_mv"] > qc_criteria["slow_noise_rms_mv_max"]:
                fail_tags.append("slow noise above threshold")

            if sweep["vm_delta_mv"] > qc_criteria["vm_delta_mv_max"]:
                fail_tags.append("Vm delta")


            # fail sweeps if stimulus duration is zero
            # Uncomment out hte following 3 lines to have sweeps without stimulus
            #   faile QC
            if sweep["stimulus_duration"] <= 0:
                desc = sweep["ephys_stimulus"]["description"]
                if not desc.startswith("EXTP"):
                    fail_tags.append("No stimulus detected")


            sweep_state[name] = {}
            if len(fail_tags) > 0:
                sweep_state[name]["state"] = "Fail"
                sweep_state[name]["reasons"] = fail_tags
            else:
                sweep_state[name]["state"] = "Pass"

        except:
            print("Error processing sweep %s" % name)
            raise

    ####################################
    # done - prepare and deliver results
    if len(exp_fail_tags) > 0:
        jout["qc_result"] = "failed"
    else:
        jout["qc_result"] = "passed"

    return jout

if __name__ == "__main__": 
    # read module input. PipelineModule object automatically parses the 
    #   command line to pull out input.json and output.json file names
    module = PipelineModule()
    jin = module.input_data()   # loads input.json
    jout = main(jin)
    module.write_output_data(jout)  # writes output.json




class Unionize(object):
    '''Abstract base class for unionize records.
    '''

    def __init__(self, *args, **kwargs):
        raise NotImplementedError()
        
        
    def calculate(self, *args, **kwargs):
        raise NotImplementedError()
        
        
    def propagate(self, ancestor, copy_all, *args, **kwargs):
        raise NotImplementedError()
        
        
    def output(self, *args, **kwargs):
        raise NotImplementedError()


    def slice_arrays(self, low, high, data_arrays):
        '''Extract a slice from several aligned arrays
        
        Parameters
        ----------
        low : int
            start of slice, inclusive
        high : int
            end of slice, exclusive
        data_arrays : dict
            keys are varieties of data. values are sorted, flattened 
            data arrays
        
        '''
    
        return {k: v[low:high] for k, v in data_arrays.iteritems()}



from __future__ import division
import logging
import sys
import functools

import numpy as np
from scipy.ndimage.interpolation import zoom
from six import iteritems

sys.path.append('/shared/bioapps/itk/itk_shared/jp2/build')
import jpeg_twok

import image_utilities as iu


#==============================================================================


class SubImage(object):

    @property
    def pixel_counter(self):
        if not hasattr(self, '_pixel_counter'):
            self._pixel_counter = self.make_pixel_counter()
        return self._pixel_counter


    def __init__(self, reduce_level, in_dims, in_spacing, coarse_spacing, 
                 *args, **kwargs):
                 
        self.reduce_level = reduce_level
        self.in_dims = np.around(in_dims / 2**reduce_level).astype(int)
        self.in_spacing = in_spacing * 2**reduce_level
        self.coarse_spacing = coarse_spacing
        
        self.blocks, self.coarse_dims = iu.grid_image_blocks(
            self.in_dims, self.in_spacing, self.coarse_spacing)
        
        self.images = {}
        self.accumulators = {}
        
        
    def setup_images(self):
        pass
        
        
    def compute_coarse_planes(self):
        raise NotImplementedError()


    def binarize(self, image_name):
        logging.info('binarizing {0}'.format(image_name))
        self.images[image_name][np.nonzero(self.images[image_name])] = 1
      

    def apply_mask(self, image_name, mask_name, positive=True):
        logging.info('applying {0} mask to {1}'.format(mask_name, image_name))
        
        mask = self.images[mask_name]
        if not positive:
            mask = np.logical_not(mask)
        mask = mask.astype(np.uint8)

        self.images[image_name] = np.multiply(self.images[image_name], mask)


    def make_pixel_counter(self):
        fn = lambda x: np.sum(x) * 2 ** ( self.reduce_level + 1) # additional x2 <- is an area
        return functools.partial(iu.block_apply, out_shape=self.coarse_dims, 
                                    dtype=np.float32, blocks=self.blocks, 
                                    fn=fn)


    def apply_pixel_counter(self, accumulator_name, image):
        self.accumulators[accumulator_name] = self.pixel_counter(image)
        
    
#==============================================================================
    

class SegmentationSubImage(SubImage):

    required_segmentations = []


    def __init__(self, reduce_level, in_dims, in_spacing, coarse_spacing, 
                 segmentation_paths, *args, **kwargs):

        super(SegmentationSubImage, self).__init__(
            reduce_level, in_dims, in_spacing, coarse_spacing, *args, **kwargs)

        self.segmentation_paths = segmentation_paths
        if 'filter_bit' in kwargs and kwargs['filter_bit'] is not None:
            self.filter = 2 ** kwargs['filter_bit']


    def setup_images(self):
        super(SegmentationSubImage, self).setup_images()
        self.get_segmentation()


    def get_segmentation(self):
        
        for name in self.__class__.required_segmentations:
            logging.info('loading {0} segmentation'.format(name))

            self.read_segmentation_image(name)

        self.process_segmentation()


    def process_segmentation(self):
        pass
    

    def extract_signal_from_segmentation(self, segmentation_name='segmentation', 
                                         signal_name='signal'):
        '''

        Notes
        -----
        Currently, the segmentation uses a series of codes to map 8-bit values 
        onto meaningful classifications. The code for signal pixels is a 1 in 
        the leftmost bit.

        In some cases, bit 5 indicates that the pixel was not removed in a 
        posfiltering process. Optionally, this postfilter can be applied in gridding.

        '''

        logging.info('extracting {0} mask'.format(signal_name))
        self.images[signal_name] = np.right_shift(self.images[segmentation_name], 7)
        
        signal_count = np.count_nonzero(self.images[signal_name])
        logging.info('{0} signal pixels were detected'.format(signal_count))

        if hasattr(self, 'filter'):
            filter_mask = np.bitwise_and(self.images[segmentation_name], self.filter)
            self.images[signal_name][filter_mask == 0] = 0

            filter_count = np.count_nonzero(self.images[signal_name])
            logging.info('{0} / {1} pixels passed the signal filter'.format(filter_count, signal_count))

            


    def extract_injection_from_segmentation(self, segmentation_name='segmentation', 
                                            injection_name='injection'):
        '''

        Notes
        -----
        Currently, the segmentation uses a series of codes to map 8-bit values 
        onto meaningful classifications. The code for signal pixels is a 1 in 
        at least one of of the 5 rightmost bits.

        '''

        logging.info('extracting {0} mask'.format(injection_name))
        self.images[injection_name] = np.bitwise_and(self.images[segmentation_name], 31)
        self.images[injection_name][self.images[injection_name] > 0] = 1


    def read_segmentation_image(self, segmentation_name='segmentation'):
        '''

        Notes
        -----
        We downsample in memory rather than using the jp2 pyramid because the 
        segmentation is a label image.

        '''
        
        logging.info('loading {0}'.format(segmentation_name))
        segmentation = jpeg_twok.read(self.segmentation_paths[segmentation_name]).T
        
        if self.reduce_level > 0:
            logging.info('downsampling {0}'.format(segmentation_name))
            segmentation = zoom(segmentation, 1.0 / 2**self.reduce_level, order=0)
            
        self.images[segmentation_name] = segmentation


#==============================================================================


class IntensitySubImage(SubImage):

    required_intensities = []


    def __init__(self, reduce_level, in_dims, in_spacing, coarse_spacing, intensity_paths, 
                 *args, **kwargs):

        super(IntensitySubImage, self).__init__(reduce_level, in_dims, 
            in_spacing, coarse_spacing, *args, **kwargs)

        self.intensity_paths = intensity_paths


    def get_intensity(self):
  
        for name in self.__class__.required_intensities:
            logging.info('loading {0}'.format(name))

            info = self.intensity_paths[name]
            self.images[name] = jpeg_twok.read(info['path'], self.reduce_level, info['channel']).T


    def setup_images(self):
        super(IntensitySubImage, self).setup_images()
        self.get_intensity()


#==============================================================================

    
class PolygonSubImage(SubImage):

    required_polys = []

    def __init__(self, reduce_level, in_dims, in_spacing, coarse_spacing, 
                 polygon_info, *args, **kwargs):
                 
        super(PolygonSubImage, self).__init__(
            reduce_level, in_dims, in_spacing, coarse_spacing, *args, **kwargs)
            
        self.polygon_info = polygon_info
        
        
    def setup_images(self):
        super(PolygonSubImage, self).setup_images()
        self.get_polygons()


    def get_polygons(self):

        for key in self.__class__.required_polys:
            logging.info('rasterizing {0} polygon'.format(key))
          
            points = self.polygon_info[key]
            self.images[key] = iu.rasterize_polygons(self.in_dims.astype(int)[::-1], 
                                                     [1.0 / 2**self.reduce_level, 
                                                      1.0 / 2**self.reduce_level], 
                                                     points).T
        

#==============================================================================
    
    
def run_subimage(input_data):
    
    # TODO: remove or fix
    logging.basicConfig(format='%(asctime)s - %(process)s - %(levelname)s - %(message)s')
    logging.getLogger('').setLevel(logging.INFO)
        
    index = input_data.pop('specimen_tissue_index')
    cls = input_data.pop('cls')
    logging.info('handling {0} at index {1}'.format(cls.__name__, index))

    si = cls(**input_data)
    
    try:
        si.setup_images()
        si.compute_coarse_planes()
    except Exception as err:
        logging.exception(err)
        raise err
    
    return index, si.accumulators



from __future__ import division
import logging
import sys
import functools

import numpy as np
from scipy.ndimage.interpolation import zoom
from six import iteritems

sys.path.append('/shared/bioapps/itk/itk_shared/jp2/build')
import jpeg_twok

import image_utilities as iu
from base_subimage import PolygonSubImage, SegmentationSubImage, IntensitySubImage


#==============================================================================


class ClassicSubImage(IntensitySubImage, SegmentationSubImage, PolygonSubImage):
    
    required_polys = ['missing_tile', 'no_signal', 'aav_exclusion']
    required_segmentations = ['segmentation']
    required_intensities = ['green']
    
    
    def __init__(self, reduce_level, in_dims, in_spacing, coarse_spacing, 
                 polygon_info, segmentation_paths, intensity_paths, 
                 *args, **kwargs):
                 
        super(ClassicSubImage, self).__init__(
            reduce_level, in_dims, in_spacing, coarse_spacing, 
            polygon_info=polygon_info, 
            segmentation_paths=segmentation_paths, 
            intensity_paths=intensity_paths, 
            *args, **kwargs)


    def process_segmentation(self):
        
        self.apply_mask('segmentation', 'missing_tile', False)
        
        self.extract_signal_from_segmentation(signal_name='projection')
        
        self.apply_mask('projection', 'no_signal', False)
        del self.images['no_signal']
            
        self.extract_injection_from_segmentation()
        del self.images['segmentation']
        
        self.binarize('projection')
        self.binarize('injection')
        
        
    def compute_coarse_planes(self):

        #  do these in batches to minimize peak memory usage
        self.compute_intensity()
        self.compute_injection()
        self.compute_projection()
        self.compute_sum_pixels()
        
        del self.images
        

    def compute_intensity(self):
        logging.info('computing green accumulators')
                
        self.apply_pixel_counter('sum_pixel_intensities', self.images['green'])

        injection_intensity = np.multiply(self.images['green'], self.images['injection'])
        self.apply_pixel_counter('injection_sum_pixel_intensities', injection_intensity)
        del injection_intensity
        
        self.images['green'][self.images['projection'] == 0] = 0
        self.apply_pixel_counter('sum_projecting_pixel_intensities', self.images['green'])
        
        self.images['green'][self.images['injection'] == 0] = 0
        self.apply_pixel_counter('injectionsum_projecting_pixel_intensities', self.images['green'])

        del self.images['green']    


    def compute_injection(self):
        logging.info('computing injection accumulators')
        
        self.apply_pixel_counter('injection_sum_pixels', self.images['injection'])

        injection_projecting_pixels = np.logical_and(self.images['injection'], self.images['projection'])
        self.apply_pixel_counter('injection_sum_projecting_pixels', injection_projecting_pixels)
        del injection_projecting_pixels

        del self.images['injection']


    def compute_projection(self):
        logging.info('computing projection accumulators')
    
        self.apply_pixel_counter('sum_projecting_pixels', self.images['projection'])
        del self.images['projection']
        
    
    def compute_sum_pixels(self):
        logging.info('computing sum pixel accumulators')

        self.apply_pixel_counter('sum_pixels', np.logical_not(self.images['missing_tile']))
        
        if 'aav_exclusion' in self.images:
            self.apply_pixel_counter('aav_exclusion_sum_pixels', self.images['aav_exclusion'])


#==============================================================================



# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import node

class Compartment(object):
    def __init__(self, node1, node2):
        if not isinstance(node1, node.Node) or not isinstance(node2, node.Node):
            raise TypeError("Must supply Node objects to Compartment constructor")
        self.length = node.euclidean_distance(node1, node2)
        self.center = node.midpoint(node1, node2)
        self.node1 = node1
        self.node2 = node2

    def __str__(self):
        s = "%s %f" % (str(self.center), self.length)
        s += "\n\t" + self.node1.short_string() + "\n\t" + self.node2.short_string()
        return s




###
# This program generates a json file containing the GBM sample metadata records from an sql query

# To Do: This will be done by the strategy and later removed from here.

import psycopg2
import json
import sys
from psycopg2.extras import RealDictCursor


def main(sample_metadata_json_location, db_host, db_port, db_name, db_user, db_passwd):

    conn = psycopg2.connect(host=db_host, port=db_port, dbname=db_name, user=db_user, password=db_passwd)
    cur = conn.cursor(cursor_factory=RealDictCursor)
    cur.execute("select distinct rna.id as rna_well_id, tumor.id as tumor_id, tumor.external_specimen_name as tumor_name"
                ", block.id as block_id, block.external_specimen_name as block_name, sp.id as specimen_id"
                ", sp.external_specimen_name as specimen_name, min(poly.id) as polygon_id, st.id as structure_id"
                ", st.acronym as structure_abbreviation, to_hex(st.red) || to_hex(st.green) || to_hex(st.blue) as "
                "structure_color, st.name as structure_name from wells rna join image_series mims on mims.id = "
                "rna.image_series_id join specimens sp on sp.id = mims.specimen_id join specimens block on block.id = "
                "sp.parent_id join specimens tumor on tumor.id = block.parent_id join avg_microarray_templates mt on "
                "mt.image_series_id = mims.id join avg_graphic_objects poly on poly.id = mt.shape_id join structures st "
                "on st.id = poly.structure_id join rs_tubes tube on tube.sample_id = rna.id join rna_seq_experiments exp "
                "on exp.rs_tube_id = tube.id join rna_seq_analysis_runs_rna_seq_experiments ar2exp on "
                "ar2exp.rna_seq_experiment_id = exp.id join analysis_runs ar on ar.id = ar2exp.rna_seq_analysis_run_id "
                "join well_known_files fpkm on fpkm.attachable_id = ar.id where rna.sample_id_string like any (array "
                "['366-___', '466-___']) and fpkm.published_at is not null group by tumor.id, "
                "tumor.external_specimen_name, block.id, block.external_specimen_name, sp.id, sp.external_specimen_name, "
                "rna.id, st.id, st.acronym, st.name, structure_color order by rna.id;")
    with open(sample_metadata_json_location, 'w') as outfile:
        json.dump(cur.fetchall(), outfile, indent=2)


if __name__ == '__main__':

    sample_metadata_json_location = sys.argv[1]
    db_host = sys.argv[2]
    db_port = sys.argv[3]
    db_name = sys.argv[4]
    db_user = sys.argv[5]
    db_passwd = sys.argv[6]
    main(sample_metadata_json_location, db_host, db_port, db_name, db_user, db_passwd)


#!/usr/bin/python
# Copyright 2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import json, sys, traceback, logging
from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module, SHARED_PYTHON
import allensdk.internal.core.lims_utilities as lu
import os
import logging

def get_experiment_nwb_file(experiment_id):
    res = lu.query("""
select * from well_known_files wkf 
join well_known_file_types wkft on wkft.id = wkf.well_known_file_type_id
where attachable_id = %d
and wkft.name = 'NWBOphys'
""" % experiment_id)
    return os.path.join(res[0]['storage_directory'], res[0]['filename'])

def get_experiment_session(experiment_id):
    return lu.query("""
select stimulus_name from ophys_sessions os
join ophys_experiments oe on oe.ophys_session_id = os.id
where oe.id = %d
""" % experiment_id)[0]['stimulus_name']

def debug(experiment_ids, local=False, 
                          conda_env=None,
                          OUTPUT_DIR = "/data/informatics/CAM/analysis/",
                          SDK_PATH = "/data/informatics/CAM/analysis/allensdk/",
                          walltime="10:00:00",
                          python=SHARED_PYTHON,
                          queue='braintv'):

    input_data = {}
    for eid in experiment_ids:
        exp_dir = os.path.join(OUTPUT_DIR, str(eid))
        input_data[eid] = dict(nwb_file=get_experiment_nwb_file(eid),
                               output_file=os.path.join(exp_dir, "%d_analysis.h5" % eid),
                               session_name=get_experiment_session(eid))



    run_module(os.path.abspath(__file__),
               input_data,
               exp_dir,
               conda_env=conda_env,
               python=python,
               sdk_path=SDK_PATH,
               pbs=dict(vmem=32,
                        job_name="bobanalysis_%d"% eid,
                        walltime=walltime,
                        queue='braintv'),
               local=local)

def main():
    
    from allensdk.brain_observatory.session_analysis import run_session_analysis
    import allensdk.brain_observatory.stimulus_info as si
    import allensdk.core.json_utilities as json_util

    mod = PipelineModule()
    jin = mod.input_data()

    results = {}

    for ident, experiment in jin.iteritems():
        nwb_file = experiment['nwb_file']
        output_file = experiment['output_file']

        if experiment["session_name"] not in si.SESSION_STIMULUS_MAP.keys():
            raise Exception("Could not run analysis for unknown session: %s" % experiment["session_name"])

        logging.info("Running %s analysis", experiment["session_name"])
        logging.info("NWB file %s", nwb_file)
        logging.info("Output file %s", output_file)

        results[ident] = run_session_analysis(nwb_file, output_file, 
                                              save_flag=True, plot_flag=False)

    logging.info("Generating output")

    jout = {}
    for session_name, data in results.items():
        # results for this session
        res = {}
        # metric fields
        names = {}
        roi_id = None
        for metric, values in data['cell'].items():
            if metric == "roi_id":
                roi_id = values
            else:
                # convert dict to array
                vals = []
                for i in range(len(values)):
                    vals.append(values[i])          # panda syntax
                names[metric] = vals
        # make an output record for each roi_id
        if roi_id is not None:
            for i in range(len(roi_id)):
                name = roi_id[i]
                roi = {}
                for field, values in names.items():
                    roi[field] = values[i]
                res[name] = roi

        jout[session_name] = {
            'cell': res,
            'experiment': data['experiment']
            }

    logging.info("Saving output")

    mod.write_output_data(jout)

if __name__ == "__main__": main()



# Copyright 2015-2016 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.

import copy
import math
import numpy as np
from node import Node
from compartment import Compartment


class Morphology( object ):
    """ 
    Keep track of the list of nodes in a morphology and provide 
    a few helper methods (soma, tree information, pruning, etc).
    """

    SOMA = 1
    AXON = 2
    BASAL_DENDRITE = 3
    APICAL_DENDRITE = 4

    NODE_TYPES = [ SOMA, AXON, BASAL_DENDRITE, APICAL_DENDRITE ]

    def __init__(self, node_list=None):
        """ 
        Try to initialize from a list of nodes first, then from
        a dictionary indexed by node id if that fails, and finally just
        leave everything empty.
        
        Parameters
        ----------
        node_list: list 
            list of Node objects
        """
        self._node_list = []            # list of morphology node IDs
        self._compartment_list = []     # list of morphology compartment IDs

        ##############################################
        # define tree list here for clarity, even though it's reset below
        #   when nodes are assigned
        self._tree_list = []

        ##############################################
        # dimensions of morphology, including min and max values on xyz
        # this is cached value
        # NOTE: if morphology is manually manipulated, this value can
        #   become incorrect
        self.dims = None
        
        ##############################################
        # construct the node list
        # first try to do so using the node list, then try using
        #   the node index and if that fails then complain
        if node_list:
            self.node_list = node_list
        ##############################################
        # verify morphology is consistent with morphology rules (e.g.,
        #   no dendrite branching from an axon)
        num_errors = self._check_consistency()
        if num_errors > 0:
            raise ValueError("Morphology appears to be inconsistent")
        ##############################################
        # restructure morphology as necessary (eg, renumber nodes)
        #   and construct internal associations
        self._reconstruct()


    ####################################################################
    ####################################################################
    # class properties, and helper functions for them

    @property 
    def node_list(self):
        """ Return the node list.  This is a property to ensure that the 
        node list and node index are in sync. """
        return self._node_list

    @node_list.setter
    def node_list(self, node_list):
        """ Update the node list.  """
        self._set_nodes(node_list)

    @property
    def compartment_list(self):
        return self._compartment_list

    @property
    def num_trees(self):
        """ Return the number of trees in the morphology. A tree is
        defined as everything following from a single root node. """
        return len(self._tree_list)

    @property
    def num_nodes(self):
        """ 
        Return the number of nodes in the morphology. 
        """
        return len(self.node_list)

    # internal function
    def _set_nodes(self, node_list):
        """
        take a list of SWC node objects and turn those into morphology
        nodes need to be able to initialize from a list supplied by an SWC
        file while also being able to initialize from the node list
        of an existing Morphology object. As nodes in a morphology object
        contain reference to nodes in that object, make a shallow copy
        of input nodes and overwrite known references (ie, the 
        'children' array)
        """
        self._node_list = []
        for obj in node_list:
            seg = copy.copy(obj)
            seg.tree_id = -1
            seg.children = []
            self._node_list.append(seg)
        # list data now set. remove holes in sequence and re-index
        self._reconstruct()

    # removed old 'soma' and 'root' calls as these were ambiguous
    # a soma can consist of multiple compartments, and there can
    #   be multiple roots
    # replaced those calls with new soma_root(), which returns the
    #   same thing as the old soma() when only a single soma node is 
    #   present
    def soma_root(self):
        """ Returns root node of soma, if present"""
        if len(self._tree_list) > 0 and self._tree_list[0][0].t == 1:
            return self._tree_list[0][0]
        return None

    ####################################################################
    ####################################################################
    # tree and node access

    def tree(self, n):
        """ 
        Returns a list of all Morphology nodes within the specified
        tree. A tree is defined as a fully connected graph of nodes.
        Each tree has exactly one root.
        
        Parameters
        ----------
        n: integer
            ID of desired tree
            
        Returns
        -------
        A list of all morphology objects in the specified tree, or None
        if the tree doesn't exist
        """
        if n < 0 or n >= len(self._tree_list):
            return None
        return self._tree_list[n]


    def node(self, n):
        """
        Returns the morphology node having the specified ID.
        
        Parameters
        ----------
        n: integer
            ID of desired node
            
        Returns
        -------
        A morphology node having the specified ID, or None if such a
        node doesn't exist
        """
        # undocumented feature -- if a node is supplied instead of a
        #   node ID, the node is returned and no error is 
        #   triggered
        return self._resolve_node_type(n)


    def compartment(self, n):
        """ 
        Returns the morphology Compartment having the specified ID.
        
        Parameters
        ----------
        n: integer
            ID of desired compartment
            
        Returns
        -------
        A morphology object having the specified ID, or None if such a
        node doesn't exist
        """
        if n < 0 or n >= len(self.compartment_list):
            return None
        return self._compartment_list[n]


    def parent_of(self, seg):
        """ Returns parent of the specified node.
        
        Parameters
        ----------
        seg: integer or Morphology Object
            The ID of the child node, or the child node itself
            
        Returns
        -------
        A morphology object, or None if no parent exists or if the
        specified node ID doesn't exist
        """
        # if ID passed in, make sure it's converted to a node
        # don't trap for exception here -- if supplied segment is 
        #   incorrect, make sure the user knows about it
        seg = self._resolve_node_type(seg)
        # return parent of specified node
        if seg is not None and seg.parent >= 0:
            return self._node_list[seg.parent]
        return None


    def children_of(self, seg):
        """ Returns a list of the children of the specified node
        
        Parameters
        ----------
        seg: integer or Morphology Object
            The ID of the parent node, or the parent node itself
            
        Returns
        -------
        A list of the child morphology objects. If the ID of the parent
        node is invalid, None is returned.
        """
        seg = self._resolve_node_type(seg)
        return [ self._node_list[c] for c in seg.children ]


    def to_dict(self):
        """
        Returns a dictionary of Node objects. These Nodes are a copy
        of the Morphology. Modifying them will not modify anything in
        the Morphology itself.
        """
        return { c.n: c.to_dict() for c in self._node_list }


    ###################################################################
    ###################################################################
    # Information querying and data manipulation

    # internal function. takes an integer and returns the node having
    #   that ID. IF a node is passed in instead, it is returned
    def _resolve_node_type(self, seg):
        # if node passed then we don't need to convert anything
        # if node not passed, try converting value to int
        #   and using that as an index
        if not isinstance(seg, Node):
            try:
                seg = int(seg)
                if seg < 0 or seg >= len(self._node_list):
                    return None
                seg = self._node_list[seg]
            except ValueError:
                raise TypeError("Object not recognized as morphology Node or index")
        return seg


    def change_parent(self, child, parent):
        """ Change the parent of a node. The child node is adjusted to 
        point to the new parent, the child is taken off of the previous 
        parent's child list, and it is added to the new parent's child list.
        
        Parameters
        ----------
        child: integer or Morphology Object
            The ID of the child node, or the child node itself
            
        parent: integer or Morphology Object
            The ID of the parent node, or the parent node itself
            
        Returns
        -------
        Nothing
        """
        child_seg = self._resolve_node_type(child)
        parent_seg = self._resolve_node_type(parent)
        # if child has former parent, remove it from parent's child list
        if child_seg.parent >= 0:
            old_par = self.node(child_seg.parent)
            old_par.children.remove(child_seg.n)
        parent_seg.children.append(child_seg.n)
        child_seg.parent = parent_seg.n
    
    def get_dimensions(self):
        """ Returns tuple of overall width, height and depth of 
            morphology. 
            WARNING: if locations of nodes in morphology are manipulated
            then this value can become incorrect. It can be reset and
            recalculated by programmitcally setting self.dims to None.

            Returns
            -------
            3 real arrays: [width, height, depth], [min_x, min_y, min_z],
            [max_x, max_y, max_z]
        """
        if self.dims is None:
            min_x = self.node_list[0].x
            max_x = self.node_list[0].x
            min_y = self.node_list[0].y
            max_y = self.node_list[0].y
            min_z = self.node_list[0].z
            max_z = self.node_list[0].z
            for node in self.node_list:
                max_x = max(node.x, max_x)
                max_y = max(node.y, max_y)
                max_z = max(node.z, max_z)
                #
                min_x = min(node.x, min_x)
                min_y = min(node.y, min_y)
                min_z = min(node.z, min_z)
            self.dims = [(max_x-min_x), (max_y-min_y), (max_z-min_z)], [min_x, min_y, min_z], [max_x, max_y, max_z]
        return self.dims

    # returns a list of node located within dist of x,y,z
    def find(self, x, y, z, dist, node_type=None):
        """ Returns a list of Morphology Objects located within 'dist' 
        of coordinate (x,y,z). If node_type is specified, the search 
        will be constrained to return only nodes of that type.
        
        Parameters
        ----------
        x, y, z: float
            The x,y,z coordinates from which to search around
        
        dist: float
            The search radius
        
        node_type: enum (optional)
            One of the following constants: SOMA, AXON, 
            BASAL_DENDRITE or APICAL_DENDRITE
            
        Returns
        -------
        A list of all Morphology Objects matching the search criteria
        """
        found = []
        for seg in self.node_list:
            dx = seg.x - x
            dy = seg.y - y
            dz = seg.z - z
            if math.sqrt(dx*dx + dy*dy + dz*dz) <= dist:
                if node_type is None or seg.t == node_type:
                    found.append(seg)
        return found


    def node_list_by_type(self, node_type):
        """ Return an list of all nodes having the specified
        node type.
        
        Parameters
        ----------
        node_type: int
            Desired node type
        
        Returns
        -------
        A list of of Morphology Objects
        """
        return [x for x in self._node_list if x.t == node_type]


    def save(self, file_name):
        """ Write this morphology out to an SWC file 
      
        Parameters
        ----------
        file_name: string
            desired name of your SWC file
        """
        f = open(file_name, "w")
        f.write("#n,type,x,y,z,radius,parent\n")
        for seg in self.node_list:
            f.write("%d %d " % (seg.n, seg.t))
            f.write("%0.4f " % seg.x)
            f.write("%0.4f " % seg.y)
            f.write("%0.4f " % seg.z)
            f.write("%0.4f " % seg.radius)
            f.write("%d\n" % seg.parent)
        f.close()


    # keep for backward compatibility, but don't publish in docs
    def write(self, file_name):
        self.save(file_name)

        
    def sparsify(self, modulo):
        """ Return a new Morphology object that has a given number of non-leaf,
        non-root nodes removed.
        
        Parameters
        ----------
        modulo: int
           keep 1 out of every modulo nodes.
        
        Returns
        -------   
        Morphology
            A new morphology instance
        """
        # create and return a new morphology instance. make a copy of
        #   this morphology's node list and manipulate that
        nodes = copy.deepcopy(self.node_list)
        # figure out which nodes to toss
        keep = {}
        ctr = 0 # mod counter -- keep every modulo element (starting w/ 1st)
        for seg in nodes:
            nid = seg.n
            if (seg.parent < 0 or 
                    len(seg.children) != 1 or 
                    nodes[seg.parent].t == Morphology.SOMA or
                    seg.t == Morphology.SOMA):
                keep[nid] = True
            else:
                if ctr % modulo == 0:
                    keep[nid] = True
                else:
                    keep[nid] = False
                ctr += 1
        # hook children up to their new parents
        for seg in nodes:
            if keep[seg.n] is False:
                parent_id = seg.parent
                while keep[parent_id] is False:
                    parent_id = nodes[parent_id].parent
                for child_id in seg.children:
                    nodes[child_id].parent = parent_id
        # filter out orphans
        sparse = []
        for seg in nodes:
            if keep[seg.n] is True:
                sparse.append(seg)
        return Morphology(sparse)


    ####################################################################
    ####################################################################
    
    def _reconstruct(self):
        """
        Internal function. 
        Restructures data and establishes appropriate internal linking. 
        Data is re-order, removing 'holes' in the ID sequence so that 
        each object ID corresponds to its position in node list. 
        Dictionaries mapping IDs to objects are no longer necessary.
        Trees are (re)calculated
        Parent-child indices are recalculated 
        A new compartment list is created
        """
        remap = {}
        # everything defaults to root. this way if a parent was deleted
        #   the child will become a new root
        for i in range(len(self.node_list)):
            remap[i] = -1
        # map old old node numbers to new ones. reset n to the new ID
        #   and put node in new list
        new_id = 0
        tmp_list = []
        for node in self.node_list:
            if node is not None:
                remap[node.n] = new_id
                node.n = new_id
                tmp_list.append(node)
                new_id += 1
        # use map to reset parent values. copy objs to new list
        for node in tmp_list:
            if node.parent >= 0:
                node.parent = remap[node.parent]
        # replace node list with newly created node list
        self._node_list = tmp_list
        # reconstruct parent/child relationship links
        ############################
        # node list is complete and sequential so don't need index
        #   to resolve relationships
        # for each node, reset children array
        # for each node, add self to parent's child list
        for node in self._node_list:
            node.children = []
            node.compartment = -1
        for node in self._node_list:
            if node.parent >= 0:
                self._node_list[node.parent].children.append(node.n)
        # update tree lists
        self._separate_trees()
        # verify that each node ID is the same as its position in the
        #   node list
        for i in range(len(self.node_list)):
            if i != self.node(i).n:
                raise RuntimeError("Internal error detected -- node list not properly formed")
        # construct compartment list
        #   (a compartment spans the distance between two nodes)
        self._compartment_list = []
        for node in self.node_list:
            node.compartment_id = -1
        for node in self.node_list:
            for child_id in node.children:
                endpoint = self.node(child_id)
                compartment = Compartment(node, endpoint)
                endpoint.compartment_id = len(self._compartment_list)
                self._compartment_list.append(compartment)


    def append(self, nodes):
        """ Add additional nodes to this Morphology. Those nodes must
        originate from another morphology object.
        
        Parameters
        ----------
        nodes: list of Morphology nodes
        """
        # construct a map between new and old IDs of added nodes
        remap = {}
        for i in range(len(nodes)):
            remap[i] = -1
        # map old old node numbers to new ones. reset n to the new ID
        # append new nodes to existing node list
        old_count = len(self.node_list)
        new_id = old_count
        for node in nodes:
            if node is not None:
                remap[node.n] = new_id
                node.n = new_id
                self._node_list.append(node)
                new_id += 1
        # use map to reset parent values. copy objs to new list
        for i in range(old_count, len(self.node_list)):
            node = self.node_list[i]
            if node.parent >= 0:
                node.parent = remap[node.parent]
        self._reconstruct()


    def convert_type(self, from_type, to_type):
        """ Convert all nodes in morphology from one type to another

        Parameters
        ----------
        from_type: enum
            The node type that will be eliminated and replaced.
            Use one of the following constants: SOMA, AXON, 
            BASAL_DENDRITE, or APICAL_DENDRITE

        to_type: enum
            The new type that will replace it. 
            Use one of the following constants: SOMA, AXON, 
            BASAL_DENDRITE, or APICAL_DENDRITE
        """
        for node in self.node_list:
            if node.t == from_type:
                node.t = to_type


    def stumpify_axon(self, count=10):
        """ Remove all axon nodes except the first 'count'
        nodes, as counted from the connected axon root.
        
        Parameters
        ----------
        count: Integer
            The length of the axon 'stump', in number of nodes
        """
        # find connected axon root
        axon_root = None
        for seg in self.node_list:
            if seg.t == Morphology.AXON:
                par_id = seg.parent
                if par_id >= 0:
                    par = self.node_list[par_id]
                    if par.t != Morphology.AXON:
                        axon_root = seg
                        break
        if axon_root is None:
            return
        # flag the first 'count' nodes from the axon root
        ax = axon_root
        for node in self.node_list:
            node.flag = None
        for i in range(count):
            # ignore bifurcations -- go 'count' deep on one line only
            ax.flag = i
            #ax["flag"] = i
            children = ax.children
            if len(children) > 0:
                ax = self.node(children[0])
        # strip out all axons that aren't flagged
        for i in range(len(self.node_list)):
            seg = self.node_list[i]
            if seg.t == Morphology.AXON:
                #if "flag" not in seg:
                if seg.flag is None:
                    self.node_list[i] = None
        self._reconstruct()
            

    def _strip(self, flagged_for_removal):
        """ Internal function with code common between 
        strip_all_other_types() and strip_type()
        """
        # if parent will be stripped and node will remain, convert
        #   parent to this type so it becomes root (otherwise root
        #   will move)
        root = self.soma_root()
        for node_id in self._node_list:
            node = self.node(node_id)
            if node.parent >= 0:
                parent = self.node(node.parent)
                parent_flag = flagged_for_removal[parent.n]
                node_flag = flagged_for_removal[node.n]
                if parent_flag and not node_flag:
                    # don't do this for soma root
                    if parent.n != root.n:
                        parent.t = node.t
                        flagged_for_removal[parent.n] = False
        # removed flagged items
        for i in range(len(self.node_list)):
            seg = self.node_list[i]
            if flagged_for_removal[seg.n]:
                # eliminate node
                self.node_list[i] = None
            elif seg.parent >= 0 and flagged_for_removal[seg.parent]:
                # parent was eliminated. make this a new root
                seg.parent = -1
        self._reconstruct()

        
    # strip out everything but the soma and the specified SWC type
    def strip_all_other_types(self, node_type, keep_soma=True):
        """ Strips everything from the morphology except for the
        specified type.
        Parent and child relationships are updated accordingly, creating
        new roots when necessary.
        
        Parameters
        ----------
        node_type: enum
            The node type to keep in the morphology. 
            Use one of the following constants: SOMA, AXON, 
            BASAL_DENDRITE, or APICAL_DENDRITE
        
        keep_soma: Boolean (optional)
            True (default) if soma nodes should remain in the 
            morpyhology, and False if the soma should also be stripped
        """
        flagged_for_removal = {}
        # scan nodes and see which ones should be removed. keep a record
        #   of them
        for seg in self.node_list:
            if seg.t == node_type:
                remove = False
            elif seg.t == 1 and keep_soma:
                remove = False
            else:
                remove = True
            if remove:
                flagged_for_removal[seg.n] = True
            else:
                flagged_for_removal[seg.n] = False
        self._strip(flagged_for_removal)


    # strip out the specified SWC type
    def strip_type(self, node_type):
        """ Strips all nodes of the specified type from the
        morphology. 
        Parent and child relationships are updated accordingly, creating
        new roots when necessary.
        
        Parameters
        ----------
        node_type: enum
            The node type to strip from the morphology.
            Use one of the following constants: SOMA, AXON, 
            BASAL_DENDRITE, or APICAL_DENDRITE
        """
        flagged_for_removal = {}
        for seg in self.node_list:
            if seg.t == node_type:
                remove = True
            else:
                remove = False
            if remove:
                flagged_for_removal[seg.n] = True
            else:
                flagged_for_removal[seg.n] = False
        self._strip(flagged_for_removal)
    

    def clone(self):
        """ Create a clone (deep copy) of this morphology
        """
        return copy.deepcopy(self)


    def apply_affine_only_rotation(self, aff):
        """ Apply an affine transform to all nodes in this 
        morphology. Only the rotation element of the transform is
        performed (i.e., although the entire transformation and 
        translation matrix is supplied, only the rotation element
        is used). The morphology is translated to the point where
        the soma root is at 0,0,0.
        
        Format of the affine matrix is:
        
        [x0 y0 z0]  [tx]
        [x1 y1 z1]  [ty]
        [x2 y2 z2]  [tz]
        
        where the left 3x3 the matrix defines the affine rotation 
        and scaling, and the right column is the translation
        vector.

        The matrix must be collapsed and stored in a list as follows:
        
        [x0 y0, z0, x1, y1, z1, x2, y2, z2, tx, ty, tz]
        
        Parameters
        ----------
        aff: 3x4 array of floats (python 2D list, or numpy 2D array)
            the transformation matrix
        """
        affine = np.copy(aff)
        # remove scale on each axis
        scale_x = abs(affine[0] + affine[3] + affine[6])
        if scale_x != 0.0:
            affine[0] /= scale_x
            affine[3] /= scale_x
            affine[6] /= scale_x
        scale_y = abs(affine[1] + affine[4] + affine[7])
        if scale_y != 0.0:
            affine[1] /= scale_y
            affine[4] /= scale_y
            affine[7] /= scale_y
        scale_z = abs(affine[2] + affine[5] + affine[8])
        if scale_z != 0.0:
            affine[2] /= scale_z
            affine[5] /= scale_z
            affine[8] /= scale_z
        # apply rotation
        for seg in self.node_list:
            x = seg.x*affine[0] + seg.y*affine[1] + seg.z*affine[2]
            y = seg.x*affine[3] + seg.y*affine[4] + seg.z*affine[5]
            z = seg.x*affine[6] + seg.y*affine[7] + seg.z*affine[8]
            seg.x = x
            seg.y = y
            seg.z = z
#        # relocate back to zero
#        soma = self.soma_root()
#        if soma is not None:
#            for seg in self.node_list:
#                seg.x -= soma.x
#                seg.y -= soma.y
#                seg.z -= soma.z


    def apply_affine(self, aff, scale=None):
        """ Apply an affine transform to all nodes in this 
        morphology. Compartment radius is adjusted as well.
        
        Format of the affine matrix is:
        
        [x0 y0 z0]  [tx]
        [x1 y1 z1]  [ty]
        [x2 y2 z2]  [tz]
        
        where the left 3x3 the matrix defines the affine rotation 
        and scaling, and the right column is the translation
        vector.
        
        The matrix must be collapsed and stored in a list as follows:
        
        [x0 y0, z0, x1, y1, z1, x2, y2, z2, tx, ty, tz]
        
        Parameters
        ----------
        aff: 3x4 array of floats (python 2D list, or numpy 2D array)
            the transformation matrix
        """
        # In addition to transforming the locations of the morphology
        #   nodes, the radius of each node must be adjusted.
        # There are 2 ways to measure scale from a transform. Assuming
        #   an isotropic transform, the scale is the cube root of the
        #   matrix determinant. The other ways is to measure scale 
        #   independently along each axis.
        # For now, the node radius is only updated based on the average
        #   scale along all 3 axes (eg, isotropic assumption), so calculate
        #   scale using the determinant
        #
        if scale is None:
            # calculate the determinant
            det0 = aff[0] * (aff[4]*aff[8] - aff[5]*aff[7])
            det1 = aff[1] * (aff[3]*aff[8] - aff[5]*aff[6])
            det2 = aff[2] * (aff[3]*aff[7] - aff[4]*aff[6])
            det = det0 + det1 + det2
            # determinant is change of volume that occurred during transform
            # assume equal scaling along all axes. take 3rd root to get
            #   scale factor
            det_scale = math.pow(abs(det), 1.0/3.0)
            ## measure scale along each axis
            ## keep this code here in case 
            #scale_x = abs(aff[0] + aff[3] + aff[6])
            #scale_y = abs(aff[1] + aff[4] + aff[7])
            #scale_z = abs(aff[2] + aff[5] + aff[8])
            #avg_scale = (scale_x + scale_y + scale_z) / 3.0;
            #
            # use determinant for scaling for now as it's most simple
            scale = det_scale
        for seg in self.node_list:
            x = seg.x*aff[0] + seg.y*aff[1] + seg.z*aff[2] + aff[9]
            y = seg.x*aff[3] + seg.y*aff[4] + seg.z*aff[5] + aff[10]
            z = seg.x*aff[6] + seg.y*aff[7] + seg.z*aff[8] + aff[11]
            seg.x = x
            seg.y = y
            seg.z = z
            seg.radius *= scale


    def _separate_trees(self):
        """
        Construct list of independent trees (each tree has a root of -1).
        The soma root, if it exists, is in tree 0.
        """
        trees = []
        # reset each node's tree ID to indicate that it's not assigned
        for seg in self.node_list:
            seg.tree_id = -1
        # construct trees for each node
        # if a node is adjacent an existing tree, merge to it
        # if a node is adjacent multiple trees, merge all
        for seg in self.node_list:
            # see what trees this node is adjacent to
            local_trees = []
            if seg.parent >= 0 and self.node_list[seg.parent].tree_id >= 0:
                local_trees.append(self.node_list[seg.parent].tree_id)
            for child_id in seg.children:
                child = self.node_list[child_id]
                if child.tree_id >= 0:
                    local_trees.append(child.tree_id)
            # figure out which tree to put node into
            # if there are muliple possibilities, merge all of them
            if len(local_trees) == 0:
                tree_num = len(trees) # create new tree
            elif len(local_trees) == 1:
                tree_num = local_trees[0]   # use existing tree
            elif len(local_trees) > 1:
                # this node is an intersection of multiple trees
                # merge all trees into the first one found
                tree_num = local_trees[0]
                for j in range(1,len(local_trees)):
                    dead_tree = local_trees[j]
                    trees[dead_tree] = []
                    for node in self.node_list:
                        if node.tree_id == dead_tree:
                            node.tree_id = tree_num
            # merge node into tree
            # ensure there's space
            while len(trees) <= tree_num:
                trees.append([])
            trees[tree_num].append(seg)
            seg.tree_id = tree_num
        # consolidate tree lists into class's tree list object
        self._tree_list = []
        for tree in trees:
            if len(tree) > 0:
                self._tree_list.append(tree)
        # make soma's tree be the first tree, if soma present
        # this should be the case if the file is properly ordered, but
        #   don't assume that
        soma_tree = -1
        for seg in self.node_list:
            if seg.t == 1:
                soma_tree = seg.tree_id
                break
        if soma_tree > 0:
            # swap soma tree for first tree in list
            tmp = self._tree_list[soma_tree]
            self._tree_list[soma_tree] = self._tree_list[0]
            self._tree_list[0] = tmp
        # reset node tree_id to correct tree number
        self._reset_tree_ids()


    def _reset_tree_ids(self):
        """
        reset each node's tree_id value to the correct tree number
        """
        for i in range(len(self._tree_list)):
            for j in range(len(self._tree_list[i])):
                self._tree_list[i][j].tree_id = i


    def _check_consistency(self):
        """
        internal function -- don't publish in the docs
        TODO? print warning if unrecognized types are present
        Return value: number of errors detected in file
        """
        errs = 0
        # Make sure that the parents are of proper ID range
        n = self.num_nodes
        for seg in self.node_list:
            if seg.parent >= 0:
                if seg.parent >= n:
                    print("Parent for node %d is invalid (%d)" % (seg.n, seg.parent))
                    errs += 1
        # make sure that each tree has exactly one root
        for i in range(self.num_trees):
            tree = self.tree(i)
            root = -1
            for j in range(len(tree)):
                if tree[j].parent == -1:
                    if root >= 0:
                        print("Too many roots in tree %d" % i)
                        errs += 1
                    root = j
            if root == -1:
                print("No root present in tree %d" % i)
                errs += 1
        # make sure each axon has at most one root
        # find type boundaries. at each axon boundary, walk back up
        #   tree to root and make sure another axon segment not
        #   encountered
        adoptees = self._find_type_boundary()
        for child in adoptees:
            if child.t == Morphology.AXON:
                par_id = child.parent
                while par_id >= 0:
                    par = self.node_list[par_id]
                    if par.t == Morphology.AXON:
                        print("Branch has multiple axon roots")
                        print(child)
                        print(par)
                        errs += 1
                        break
                    par_id = par.parent
        if errs > 0:
            print("Failed consistency check: %d errors encountered" % errs)
        return errs


    def _find_type_boundary(self):
        """
        return a list of segments who have parents that are a different type
        """
        adoptees = []
        for node in self.node_list:
            par = self.parent_of(node)
            if par is None:
                continue
            if node.t != par.t:
                adoptees.append(node)
        return adoptees


    # remove tree from swc's "forest"
    def delete_tree(self, n):
        """ Delete tree, and all of its nodes, from the morphology.
        
        Parameters
        ----------
        n: Integer
            The tree number to delete
        """
        if n < 0:
            return
        if n >= self.num_trees:
            print("Error -- attempted to delete non-existing tree (%d)" % n)
            raise ValueError
        tree = self.tree(n)
        for i in range(len(tree)):
            self.node_list[tree[i].n] = None
        del self._tree_list[n]
        self._reconstruct()
        # reset node tree_id to correct tree number
        self._reset_tree_ids()


    def _print_all_nodes(self):
        """
        debugging function. prints all nodes
        """
        for node in self.node_list:
            print(node.short_string())




import numpy as np

def get_sweep_v_i_t_from_set(data_set, sweep_number):
    sweep_data = data_set.get_sweep(sweep_number)
    i = sweep_data["stimulus"] # in A
    v = sweep_data["response"] # in V
    i *= 1e12 # to pA
    v *= 1e3 # to mV
    sampling_rate = sweep_data["sampling_rate"] # in Hz
    t = np.arange(0, len(v)) * (1.0 / sampling_rate)
    return v, i, t

def get_sweeps_of_type(sweep_type, sweeps):
    sweeps = [ s for s in sweeps if s['ephys_stimulus']['description'].startswith( sweep_type )]
    sweep_numbers = [ s['sweep_number'] for s in sweeps ]
    statuses = [ s['workflow_state'] for s in sweeps ]
    
    return sweeps, sweep_numbers, statuses

def get_step_stim_characteristics(i, t):
    # Assumes that there is a test pulse followed by the stimulus step
    di = np.diff(i)
    up_idx = np.flatnonzero(di > 0)
    down_idx = np.flatnonzero(di < 0)

    # second step is the stimulus
    if len(up_idx) < 2 or len(down_idx) < 2:
        return (np.nan, np.nan, np.nan, np.nan, np.nan)

    if up_idx[1] < down_idx[1]: # positive step
        start_idx = up_idx[1] + 1 # shift by one to compensate for diff()
        end_idx = down_idx[1] + 1
    else: # negative step
        start_idx = down_idx[1] + 1
        end_idx = up_idx[1] + 1
    stim_start = float(t[start_idx])
    stim_dur = float(t[end_idx] - t[start_idx])
    stim_amp = float(i[start_idx])
    return (stim_start, stim_dur, stim_amp, start_idx, end_idx)



import sys, os, shutil
import logging
from collections import defaultdict
import numpy as np
import json

from allensdk.config.manifest import Manifest
from allensdk.core.json_utilities import json_handler

from allensdk.core.nwb_data_set import NwbDataSet
from allensdk.ephys.extract_cell_features import extract_cell_features, extract_sweep_features
from allensdk.ephys.ephys_features import FeatureError
from allensdk.ephys.ephys_extractor import reset_long_squares_start
import allensdk.internal.ephys.plot_qc_figures as plot_qc_figures


TEST_PULSE_DURATION_SEC = 0.4

LONG_SQUARE_COARSE = 'C1LSCOARSE'
LONG_SQUARE_FINE = 'C1LSFINEST'
SHORT_SQUARE = 'C1SSFINEST'
RAMP = 'C1RP25PR1S'
PASSED_SWEEP_STATES = [ 'manual_passed', 'auto_passed' ]
ICLAMP_UNITS = [ 'Amps', 'pA' ]

def filter_sweeps(sweeps, types=None, passed_only=True, iclamp_only=True):
    if passed_only:
        sweeps = [ s for s in sweeps if s.get('workflow_state', None) in PASSED_SWEEP_STATES ]

    if iclamp_only:
        sweeps = [ s for s in sweeps if s['stimulus_units'] in ICLAMP_UNITS ]

    if types:
        sweeps = [ s for s in sweeps for t in types 
                   if s['ephys_stimulus']['description'].startswith(t) ]
        
    return sorted(sweeps, key=lambda x: x['sweep_number'])

def filtered_sweep_numbers(sweeps, types=None, passed_only=True, iclamp_only=True):
    return [ s['sweep_number'] for s in filter_sweeps(sweeps, types, passed_only, iclamp_only) ]

def find_stim_start(stim, idx0=0):
    """ 
    Find the index of the first nonzero positive or negative jump in an array. 

    Parameters
    ----------
    stim: np.ndarray
        Array to be searched

    idx0: int
        Start searching with this index (default: 0).

    Returns
    -------
    int
    """

    di = np.diff(stim)
    idxs = np.flatnonzero(di)
    idxs = idxs[idxs >= idx0]
    
    if len(idxs) == 0:
        return -1
    
    return idxs[0]+1

def find_sweep_stim_start(data_set, sweep_number):
    sweep = data_set.get_sweep(sweep_number)
    sr = sweep['sampling_rate']
    stim_start = find_stim_start(sweep['stimulus'], TEST_PULSE_DURATION_SEC * sr) / sr
    logging.info("Long square stims start at time %f", stim_start)
    return stim_start
 
def find_coarse_long_square_amp_delta(sweeps, decimals=0):
    """ Find the delta between amplitudes of coarse long square sweeps.  Includes failed sweeps.  """
    sweeps = filter_sweeps(sweeps, types=[ LONG_SQUARE_COARSE ], passed_only = False)
    amps = sorted([ s['stimulus_amplitude'] for s in sweeps ])
    amp_diff = np.round(np.diff(amps), decimals=decimals)

    # repeats are okay!
    amp_diff_no_zero = list(set(np.unique(amp_diff)) - set([0]))

    if len(amp_diff_no_zero) == 0:
        return 0
    
    if len(amp_diff_no_zero) != 1:
        raise FeatureError("Long square sweeps must have even amplitude step differences.  Found %s" % str(amp_diff_no_zero))

    return amp_diff_no_zero[0]

def update_output_sweep_features(cell_features, sweep_features, sweep_index):
    # add peak deflection for subthreshold long squares
    for sweep_number, sweep in sweep_index.iteritems():
        pd = sweep_features.get(sweep_number,{}).get('peak_deflect', None)
        if pd is not None:
            sweep['peak_deflection'] = pd[0]
        
    # update num_spikes
    for sweep_num in sweep_features:
        num_spikes = len(sweep_features[sweep_num]['spikes'])
        if num_spikes == 0:
            num_spikes = None
        sweep_index[sweep_num]['num_spikes'] = num_spikes


def nan_get(obj, key):
    """ Return a value from a dictionary.  If it does not exist, return None.  If it is NaN, return None """
    v = obj.get(key, None)

    if v is None:
        return None
    else:
        return None if np.isnan(v) else v

def generate_output_cell_features(cell_features, sweep_features, sweep_index):
    ephys_features = {}

    # find hero and rheo sweeps in sweep table
    rheo_sweep_num = cell_features["long_squares"]["rheobase_sweep"]["id"]
    rheo_sweep_id = sweep_index.get(rheo_sweep_num, {}).get('id', None)

    if rheo_sweep_id is None:
        raise Exception("Could not find id of rheobase sweep number %d." % rheo_sweep_num)

    hero_sweep = cell_features["long_squares"]["hero_sweep"]
    if hero_sweep is None:
        raise Exception("Could not find hero sweep")
    
    hero_sweep_num = hero_sweep["id"]
    hero_sweep_id = sweep_index.get(hero_sweep_num, {}).get('id', None)

    if hero_sweep_id is None:
        raise Exception("Could not find id of hero sweep number %d." % hero_sweep_num)
    
    # create a table of values 
    # this is a dictionary of ephys_features
    base = cell_features["long_squares"]
    ephys_features["rheobase_sweep_id"] = rheo_sweep_id
    ephys_features["rheobase_sweep_num"] = rheo_sweep_num
    ephys_features["thumbnail_sweep_id"] = hero_sweep_id
    ephys_features["thumbnail_sweep_num"] = hero_sweep_num
    ephys_features["vrest"] = nan_get(base, "v_baseline")
    ephys_features["ri"] = nan_get(base, "input_resistance")

    # change the base to hero sweep
    base = cell_features["long_squares"]["hero_sweep"]
    ephys_features["adaptation"] = nan_get(base, "adapt")
    ephys_features["latency"] = nan_get(base, "latency")

    # convert to ms
    mean_isi = nan_get(base, "mean_isi")
    ephys_features["avg_isi"] = (mean_isi * 1e3) if mean_isi is not None else None

    # now grab the rheo spike
    base = cell_features["long_squares"]["rheobase_sweep"]["spikes"][0]
    ephys_features["upstroke_downstroke_ratio_long_square"] = nan_get(base, "upstroke_downstroke_ratio") 
    ephys_features["peak_v_long_square"] = nan_get(base, "peak_v")
    ephys_features["peak_t_long_square"] = nan_get(base, "peak_t")
    ephys_features["trough_v_long_square"] = nan_get(base, "trough_v")
    ephys_features["trough_t_long_square"] = nan_get(base, "trough_t")
    ephys_features["fast_trough_v_long_square"] = nan_get(base, "fast_trough_v")
    ephys_features["fast_trough_t_long_square"] = nan_get(base, "fast_trough_t")
    ephys_features["slow_trough_v_long_square"] = nan_get(base, "slow_trough_v")
    ephys_features["slow_trough_t_long_square"] = nan_get(base, "slow_trough_t")
    ephys_features["threshold_v_long_square"] = nan_get(base, "threshold_v")
    ephys_features["threshold_i_long_square"] = nan_get(base, "threshold_i")
    ephys_features["threshold_t_long_square"] = nan_get(base, "threshold_t")
    ephys_features["peak_v_long_square"] = nan_get(base, "peak_v")
    ephys_features["peak_t_long_square"] = nan_get(base, "peak_t")

    base = cell_features["long_squares"]
    ephys_features["sag"] = nan_get(base, "sag")
    # convert to ms
    tau = nan_get(base, "tau")
    ephys_features["tau"] = (tau * 1e3) if tau is not None else None 
    ephys_features["vm_for_sag"] = nan_get(base, "vm_for_sag")
    ephys_features["has_burst"] = None#base.get("has_burst", None)
    ephys_features["has_pause"] = None#base.get("has_pause", None)
    ephys_features["has_delay"] = None#base.get("has_delay", None)
    ephys_features["f_i_curve_slope"] = nan_get(base, "fi_fit_slope")

    # change the base to ramp
    base = cell_features["ramps"]["mean_spike_0"] # mean feature of first spike for all of these
    ephys_features["upstroke_downstroke_ratio_ramp"] = nan_get(base, "upstroke_downstroke_ratio")
    ephys_features["peak_v_ramp"] = nan_get(base, "peak_v")
    ephys_features["peak_t_ramp"] = nan_get(base, "peak_t")
    ephys_features["trough_v_ramp"] = nan_get(base, "trough_v")
    ephys_features["trough_t_ramp"] = nan_get(base, "trough_t")
    ephys_features["fast_trough_v_ramp"] = nan_get(base, "fast_trough_v")
    ephys_features["fast_trough_t_ramp"] = nan_get(base, "fast_trough_t")
    ephys_features["slow_trough_v_ramp"] = nan_get(base, "slow_trough_v")
    ephys_features["slow_trough_t_ramp"] = nan_get(base, "slow_trough_t")

    ephys_features["threshold_v_ramp"] = nan_get(base, "threshold_v")
    ephys_features["threshold_i_ramp"] = nan_get(base, "threshold_i")
    ephys_features["threshold_t_ramp"] = nan_get(base, "threshold_t")

    # change the base to short_square
    base = cell_features["short_squares"]["mean_spike_0"] # mean feature of first spike for all of these
    ephys_features["upstroke_downstroke_ratio_short_square"] = nan_get(base, "upstroke_downstroke_ratio")
    ephys_features["peak_v_short_square"] = nan_get(base, "peak_v")
    ephys_features["peak_t_short_square"] = nan_get(base, "peak_t")

    ephys_features["trough_v_short_square"] = nan_get(base, "trough_v")
    ephys_features["trough_t_short_square"] = nan_get(base, "trough_t")

    ephys_features["fast_trough_v_short_square"] = nan_get(base, "fast_trough_v")
    ephys_features["fast_trough_t_short_square"] = nan_get(base, "fast_trough_t")

    ephys_features["slow_trough_v_short_square"] = nan_get(base, "slow_trough_v")
    ephys_features["slow_trough_t_short_square"] = nan_get(base, "slow_trough_t")

    ephys_features["threshold_v_short_square"] = nan_get(base, "threshold_v")
    #ephys_features["threshold_i_short_square"] = nan_get(base, "threshold_i")
    ephys_features["threshold_t_short_square"] = nan_get(base, "threshold_t")

    ephys_features["threshold_i_short_square"] = nan_get(cell_features["short_squares"], "stimulus_amplitude")

    return ephys_features

def extract_data(data, nwb_file):
    ##########################################################
    #### alings with ephys_sweep_qc_tool extract_features ####
    cell_specimen = data['specimens'][0]
    sweep_list = cell_specimen['ephys_sweeps']
    sweep_index = { s['sweep_number']:s for s in sweep_list }

    data_set = NwbDataSet(nwb_file)


    # extract sweep-level features
    logging.debug("Computing sweep features")
    iclamp_sweep_list = filter_sweeps(sweep_list, iclamp_only=True, passed_only=False)
    iclamp_sweeps = defaultdict(list)
    for s in iclamp_sweep_list:
        try:
            stimulus_type_name = s['ephys_stimulus']['ephys_stimulus_type']['name']
        except KeyError as e:
            raise Exception("Sweep %d has no ephys stimulus record in features JSON file: %s" % (s['sweep_number'], json.dumps(s, indent=3, default=json_handler)))

        if stimulus_type_name == "Unknown":
            raise Exception(("Sweep %d (%s) has 'Unknown' stimulus type." +
                             "Please update the EpysStimuli and EphysRawStimulusNames associations in LIMS.") % (s['sweep_number'], s['ephys_stimulus']['description']))

        iclamp_sweeps[stimulus_type_name].append(s['sweep_number'])

    passed_iclamp_sweep_list = filter_sweeps(sweep_list, iclamp_only=True, passed_only=True)
    num_passed_sweeps = len(passed_iclamp_sweep_list)
    logging.info("%d of %d sweeps passed QC", 
                 num_passed_sweeps, 
                 len(iclamp_sweep_list))

    if num_passed_sweeps == 0:
        raise FeatureError("There are no QC-passed sweeps available to analyze")

    # compute sweep features
    logging.info("Computing sweep features")
    sweep_features = extract_sweep_features(data_set, iclamp_sweeps)
    cell_specimen['sweep_ephys_features'] = sweep_features

    # extract cell-level features
    logging.info("Computing cell features")
    long_square_sweep_numbers = filtered_sweep_numbers(sweep_list, [ LONG_SQUARE_COARSE, LONG_SQUARE_FINE ])
    short_square_sweep_numbers = filtered_sweep_numbers(sweep_list, [ SHORT_SQUARE ])
    ramp_sweep_numbers = filtered_sweep_numbers(sweep_list, [ RAMP ])

    logging.debug("long square sweeps: %s", str(long_square_sweep_numbers))
    logging.debug("short square sweeps: %s", str(short_square_sweep_numbers))
    logging.debug("ramp sweeps: %s", str(ramp_sweep_numbers))

    # PBS-262 -- have variable subthreshold minimum for human cells
    subthresh_min_amp = None    # None means default (mouse) behavior
    long_square_amp_delta = find_coarse_long_square_amp_delta(sweep_list)
    
    if long_square_amp_delta != 20.0:
        subthresh_min_amp = -200

    logging.info("Long squares using %fpA step size.  Using subthreshold minimum amplitude of %s.", 
                 long_square_amp_delta, 
                 str(subthresh_min_amp) if subthresh_min_amp is not None else "[default]")

    stim_start = find_sweep_stim_start(data_set, long_square_sweep_numbers[0])
    if stim_start > 0:
        logging.info("resetting long square start time to: %f", stim_start)
        reset_long_squares_start(stim_start)

    cell_features = extract_cell_features(data_set, 
                                          ramp_sweep_numbers,
                                          short_square_sweep_numbers,
                                          long_square_sweep_numbers,
                                          subthresh_min_amp)
    # shuffle peak deflection for the subthreshold long squares
    for s in cell_features["long_squares"]["subthreshold_sweeps"]:
        sweep_features[s['id']]['peak_deflect'] = s['peak_deflect']

    cell_specimen['cell_ephys_features'] = cell_features

    update_output_sweep_features(cell_features, sweep_features, sweep_index)
    ephys_features = generate_output_cell_features(cell_features, sweep_features, sweep_index)

    try:
        out_ephys_features = cell_specimen.get('ephys_features',[])[0]
        out_ephys_features.update(ephys_features)
    except IndexError:
        cell_specimen['ephys_features'] = [ ephys_features ]

    #### breaks with ephys_sweep_qc_tool extract_features ####
    ##########################################################
    return sweep_list, sweep_features

def save_qc_figures(qc_fig_dir, nwb_file, output_data, plot_cell_figures):
    if os.path.exists(qc_fig_dir):
        logging.warning("removing existing qc figures directory: %s", qc_fig_dir)
        shutil.rmtree(qc_fig_dir)

    Manifest.safe_mkdir(qc_fig_dir)

    logging.debug("saving qc plot figures")
    plot_qc_figures.make_sweep_page(nwb_file, output_data, qc_fig_dir)
    plot_qc_figures.make_cell_page(nwb_file, output_data, qc_fig_dir, plot_cell_figures)



from mpi4py import MPI # needed for NEURON parallel execution
import os
from allensdk.internal.model.biophysical.deap_utils import Utils
import neuron_parallel
import logging
import logging.config as lc
import argparse
import random
import numpy as np
from deap import algorithms, base, creator, tools
from allensdk.model.biophys_sim.config import Config
from pkg_resources import resource_filename #@UnresolvedImport


BOUND_LOWER, BOUND_UPPER = 0.0, 1.0
DEFAULT_NGEN = 500
DEFAULT_MU = 1200


_optimize_log = logging.getLogger('allensdk.model.biophysical.optimize')


utils = None
h = None
do_block_check = None
t_vec = None
v_ved = None
i_vec = None
stim_params = None
max_stim_amp = None
config = None
seed = None


def eval_param_set(params):
    utils.set_normalized_parameters(params)
    h.finitialize()
    h.run()
    feature_errors = utils.calculate_feature_errors(t_vec.as_numpy(), v_vec.as_numpy(), i_vec.as_numpy())
    min_fail_penalty = 75.0
    if do_block_check and np.sum(feature_errors) < min_fail_penalty * len(feature_errors):
        if check_for_block():
            feature_errors = min_fail_penalty * np.ones_like(feature_errors)
        # Reset the stimulus back
        utils.set_iclamp_params(stim_params["amplitude"], stim_params["delay"],
            stim_params["duration"])

    return [np.sum(feature_errors)]


def check_for_block():
    utils.set_iclamp_params(max_stim_amp, stim_params["delay"],
        stim_params["duration"])
    h.finitialize()
    h.run()

    v = v_vec.as_numpy()
    t = t_vec.as_numpy()
    stim_start_idx = np.flatnonzero(t >= utils.stim.delay)[0]
    stim_end_idx = np.flatnonzero(t >= utils.stim.delay + utils.stim.dur)[0]
    depol_block_threshold = -50.0 # mV
    block_min_duration = 50.0 # ms
    long_hyperpol_threshold = -75.0 # mV

    bool_v = np.array(v > depol_block_threshold, dtype=int)
    up_indexes = np.flatnonzero(np.diff(bool_v) == 1)
    down_indexes = np.flatnonzero(np.diff(bool_v) == -1)
    if len(up_indexes) > len(down_indexes):
        down_indexes = np.append(down_indexes, [stim_end_idx])

    if len(up_indexes) == 0:
        # if it never gets high enough, that's not a good sign (meaning no spikes)
        return True
    else:
        max_depol_duration = np.max([t[down_indexes[k]] - t[up_idx] for k, up_idx in enumerate(up_indexes)])
        if max_depol_duration > block_min_duration:
            return True

    bool_v = np.array(v > long_hyperpol_threshold, dtype=int)
    up_indexes = np.flatnonzero(np.diff(bool_v) == 1)
    down_indexes = np.flatnonzero(np.diff(bool_v) == -1)
    down_indexes = down_indexes[(down_indexes > stim_start_idx) & (down_indexes < stim_end_idx)]
    if len(down_indexes) != 0:
        up_indexes = up_indexes[(up_indexes > stim_start_idx) & (up_indexes < stim_end_idx) & (up_indexes > down_indexes[0])]
        if len(up_indexes) < len(down_indexes):
            up_indexes = np.append(up_indexes, [stim_end_idx])
        max_hyperpol_duration = np.max([t[up_indexes[k]] - t[down_idx] for k, down_idx in enumerate(down_indexes)])
        if max_hyperpol_duration > block_min_duration:
            return True
    return False


def uniform(lower, upper, size=None):
    if size is None:
        return [random.uniform(a, b) for a, b in zip(lower, upper)]
    else:
        return [random.uniform(a, b) for a, b in zip([lower] * size, [upper] * size)]


def best_sum(d):
    return np.sum(d, axis=1).min()


def initPopulation(pcls, ind_init, popfile):
    popdata = np.loadtxt(popfile)
    return pcls(ind_init(utils.normalize_actual_parameters(line)) for line in popdata.tolist())


def main():
    global utils, h, v_vec, i_vec, t_vec, do_block_check, max_stim_amp, stim_params, config, seed
    parser = argparse.ArgumentParser(description='Start a DEAP testing run.')
    parser.add_argument('seed', type=int)
    parser.add_argument('config_path')
    parser.add_argument('ngen', type=int)
    parser.add_argument('mu', type=int)
    args = parser.parse_args()
    seed = args.seed

    # Set up NEURON
    config = Config().load(args.config_path)

    if 'LOG_CFG' in os.environ:
        log_config = os.environ['LOG_CFG']
    else:
        log_config = resource_filename('allensdk.model.biophysical',
                                       'logging.conf')
        os.environ['LOG_CFG'] = log_config
    lc.fileConfig(log_config)

    stim_params = config.data["stimulus"][0]

    block_check_fit_types = ["f9", "f13"]
    do_block_check = False
    if config.data["fit_name"] in block_check_fit_types:
        max_stim_amp = config.data["fitting"][0]["max_stim_test_na"]
        if max_stim_amp > stim_params["amplitude"]:
            _optimize_log.debug("Will check for blocks")
            do_block_check = True

    utils = Utils(config)
    h = utils.h

    manifest = config.manifest
    morphology_path = manifest.get_path('MORPHOLOGY')
    utils.generate_morphology(morphology_path.encode('ascii', 'ignore'))
    utils.load_cell_parameters()
    utils.insert_iclamp()
    utils.set_iclamp_params(stim_params["amplitude"], stim_params["delay"],
        stim_params["duration"])

    h.tstop = stim_params["delay"] * 2.0 + stim_params["duration"]
    h.cvode_active(1)
    h.cvode.atolscale("cai", 1e-4)
    h.cvode.maxstep(10)

    v_vec, i_vec, t_vec = utils.record_values()

    try: # Wrapping this all to catch exceptions during NEURON parallel execution
        neuron_parallel.runworker()

        # Set up genetic algorithm

        _optimize_log.debug("Setting up genetic algorithm")
        random.seed(seed)

        ngen = args.ngen
        mu = args.mu
        cxpb = 0.1
        mtpb = 0.35
        eta = 10.0

        ndim = len(config.data["channels"]) + len(config.data["addl_params"])

        creator.create("FitnessMin", base.Fitness, weights=(-1.0, ))
        creator.create("Individual", list, fitness=creator.FitnessMin)

        toolbox = base.Toolbox()

        toolbox.register("attr_float", uniform, BOUND_LOWER, BOUND_UPPER, ndim)
        toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.attr_float)
        toolbox.register("population", tools.initRepeat, list, toolbox.individual)

        toolbox.register("evaluate", eval_param_set)
        toolbox.register("mate", tools.cxSimulatedBinaryBounded, low=BOUND_LOWER, up=BOUND_UPPER,
            eta=eta)
        toolbox.register("mutate", tools.mutPolynomialBounded, low=BOUND_LOWER, up=BOUND_UPPER,
            eta=eta, indpb=mtpb)
        toolbox.register("variate", algorithms.varAnd)
        toolbox.register("select", tools.selBest)
        toolbox.register("map", neuron_parallel.map)

        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("min", np.min, axis=0)
        stats.register("max", np.max, axis=0)
        stats.register("best", best_sum)

        logbook = tools.Logbook()
        logbook.header = "gen", "nevals", "min", "max", "best"

        if "STARTPOP" in manifest.path_info:
            _optimize_log.debug("Using a pre-defined starting population")
            start_pop_path = config.manifest.get_path("STARTPOP")
            toolbox.register("population_start", initPopulation, list, creator.Individual)
            pop = toolbox.population_start(start_pop_path)
        else:
            pop = toolbox.population(n=mu)

        invalid_ind = [ind for ind in pop if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        hof = tools.HallOfFame(mu)
        hof.update(pop)

        record = stats.compile(pop)
        logbook.record(gen=0, nevals=len(invalid_ind), **record)
        _optimize_log.debug(logbook.stream)

        for gen in range(1, ngen + 1):
            offspring = toolbox.variate(pop, toolbox, cxpb, 1.0)

            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit

            hof.update(offspring)

            pop[:] = toolbox.select(pop + offspring, mu)

            record = stats.compile(pop)
            logbook.record(gen=gen, nevals=len(invalid_ind), **record)
            _optimize_log.debug(logbook.stream)

        fit_dir = config.manifest.get_path("FITDIR")
        seed_dir = fit_dir + "/s{:d}/".format(seed)
        np.savetxt(seed_dir + "final_pop.txt", np.array(map(utils.actual_parameters_from_normalized, pop)))
        np.savetxt(seed_dir + "final_pop_fit.txt", np.array([ind.fitness.values for ind in pop]))
        np.savetxt(seed_dir + "final_hof.txt", np.array(map(utils.actual_parameters_from_normalized, hof)))
        np.savetxt(seed_dir + "final_hof_fit.txt", np.array([ind.fitness.values for ind in hof]))
        neuron_parallel.done()
        h.quit()
    except RuntimeError:
        _optimize_log.critical("Exception encountered during parallel NEURON execution")
        MPI.COMM_WORLD.Abort() # Shut down all the processes


if __name__ == "__main__":
    main()



from neuron import h
import logging

_neuron_parallel_log = logging.getLogger('allensdk.model.biophysical.neuron_parallel')

_pc = h.ParallelContext()

def map(func, *iterables):
    start_time = pc_time()
    userids = []
    userid = 200 # arbitrary, but needs to be a positive integer
    for args in zip(*iterables):
        args2 = (list(a) for a in args)
        _pc.submit(userid, func, *args2)
        userids.append(userid)
        userid += 1
    results = dict(working())
    end_time = pc_time()
    _neuron_parallel_log.debug("Map took %s" % (str(end_time - start_time)))
    return [results[userid] for userid in userids]

def working():
    while _pc.working():
        userid = int(_pc.userid())
        ret = _pc.pyret()
        yield userid, ret

def runworker():
    _pc.runworker()

def done():
    _pc.done()

def pc_time():
    return _pc.time()

def reset_neuron_library():
    '''
    See Also: https://www.neuron.yale.edu/phpBB/viewtopic.php?f=2&t=2367
    '''
    _pc.gid_clear()
    
    for sec in h.allsec():
        h("%s{delete_section()}" % (sec.name()) )
    
    


from __future__ import division
import logging

from allensdk.core.simple_tree import SimpleTree

from run_tissuecyte_unionize_classic import get_ancestor_id_map, get_volume_scale
from allensdk.internal.mouse_connectivity.interval_unionize.cav_unionizer import CavUnionizer
import data_utilities as du


def run(input_data):

    logging.info('making ancestor id map')
    ancestor_id_map = get_ancestor_id_map(input_data['structures'])

    logging.info('computing volume scale factor')
    volume_scale = (input_data['reference_spacing'] / 10 ** 3) ** 3 # mum3 -> mm3
    logging.info('volume scale factor : {0}'.format(volume_scale))

    logging.info('reference shape : {0}'.format(input_data['reference_shape']))
    logging.info('reference spacing : {0}'.format(input_data['reference_spacing']))
    logging.info('image_series_id : {0}'.format(input_data['image_series_id']))

    annotation = du.load_annotation(input_data['annotation_path'], input_data['grid_paths']['data_mask'])

    unionizer = CavUnionizer()
    unionizer.setup_interval_map(annotation)
    del annotation

    signal_arrays = du.get_cav_density(input_data['grid_paths']['cav_density'])
    signal_arrays.update(du.get_sum_pixels(input_data['grid_paths']['sum_pixels']))

    for k, v in signal_arrays.iteritems():
        logging.info('sorting {0} array'.format(k))
        signal_arrays[k] = v.flat[unionizer.sort]
    
    logging.info('computing unionizes from directly annotated voxels')
    raw_unionizes = unionizer.direct_unionize(signal_arrays, pre_sorted=True)
    
    logging.info('propagating data to ancestor structures')
    raw_unionizes = CavUnionizer.propagate_unionizes(raw_unionizes, ancestor_id_map)

    logging.info('propagating data to bilateral unionizes')
    bilateral = CavUnionizer.propagate_to_bilateral(raw_unionizes)

    cooked_unionizes = unionizer.postprocess_unionizes(raw_unionizes, 
                                                       image_series_id=input_data['image_series_id'], 
                                                       volume_scale=volume_scale)
    cooked_bilateral = unionizer.postprocess_unionizes(bilateral, 
                                                       image_series_id=input_data['image_series_id'], 
                                                       volume_scale=volume_scale)

    for item in cooked_bilateral:
        item['hemisphere'] = '(none)'
        cooked_unionizes.append(item)

    logging.info('computed {0} unionize records'.format(len(cooked_unionizes)))
    return cooked_unionizes



###
# This program generates a json file containing the GBM analysis run records from an sql query

# To Do: This will be done by the strategy and later removed from here.

import psycopg2
import json
import sys


def main(analysis_records_json_location, db_host, db_port, db_name, db_user, db_passwd):

    conn = psycopg2.connect(host=db_host, port=db_port, dbname=db_name, user=db_user, password=db_passwd)
    cur = conn.cursor()
    cur.execute("select distinct rna.id as rna_well_id, gen.storage_directory || gen.filename, trans.storage_directory "
                "|| trans.filename from wells rna join rs_tubes t on t.sample_id = rna.id join rna_seq_experiments e "
                "on e.rs_tube_id = t.id join rna_seq_analysis_runs_rna_seq_experiments ar2e on ar2e.rna_seq_experiment_id "
                "= e.id join well_known_files gen on gen.attachable_id = ar2e.rna_seq_analysis_run_id and "
                "gen.well_known_file_type_id = 267380639 join well_known_files trans on trans.attachable_id "
                "= ar2e.rna_seq_analysis_run_id and trans.well_known_file_type_id = 267380638 where gen.published_at is "
                "not null and gen.storage_directory ilike '%/gbm/%' order by rna.id;")
    data = cur.fetchall()
    analysis_run_records = {"analysis_run_records": []}
    for item in data:
        record = {"rna_well_id": item[0], "analysis_run_gene_path": item[1], "analysis_run_transcript_path": item[2]}
        analysis_run_records["analysis_run_records"].append(record)
    with open(analysis_records_json_location, 'w') as outfile:
        json.dump(analysis_run_records, outfile)


if __name__ == '__main__':

    analysis_records_json_location = sys.argv[1]
    db_host = sys.argv[2]
    db_port = sys.argv[3]
    db_name = sys.argv[4]
    db_user = sys.argv[5]
    db_passwd = sys.argv[6]
    main(analysis_records_json_location, db_host, db_port, db_name, db_user, db_passwd)



import logging
import operator as op
from collections import defaultdict

import numpy as np



class Stitcher(object):


    def __init__(self, image_dimensions, tiles, average_tiles, channels):
        
          logging.info('image_dimensions: {0}'.format(image_dimensions))
          self.image_dimensions = image_dimensions

          self.average_tiles = defaultdict(lambda *a, **k: None, average_tiles)

          self.tiles = tiles
          self.channels = channels


    def run(self, cb=np.array):

        slice_image, stitched_indicator = initialize_images(self.image_dimensions, len(self.channels))
        missing_tiles = {}

        for tile in self.tiles:
            
            if tile.is_missing:

                missing_tiles[tile.index] = tile.get_missing_path()
                tile.initialize_image()

            else:

              tile.apply_average_tile_to_self(self.average_tiles[tile.channel])
              tile.trim_self()

            self.stitch(slice_image, stitched_indicator, tile, cb)

        return slice_image, missing_tiles


    def stitch(self, slice_image, stitched_indicator, tile, cb=np.array):

        region = tile.get_image_region()
        
        current_region = slice_image[region]
        indicator_region = stitched_indicator[region]

        stup = (tile.size['row'], tile.size['column'])
        blend = get_blend(indicator_region, stup, cb)
        blend = make_blended_tile(blend, tile.image, current_region)
        logging.info('obtained blend')

        slice_image[region] = blend
        stitched_indicator[region] = 1
        logging.info('updated image region with tile data')


def initialize_image(dimensions, nchannels, dtype, order='C'):
    return np.zeros((dimensions['row'], dimensions['column'], nchannels), dtype=dtype, order=order)


def initialize_images(dimensions, nchannels):
    return initialize_image(dimensions, nchannels, np.uint16), initialize_image(dimensions, nchannels, np.int8)


def make_blended_tile(blend, tile, current_region):
    return np.multiply((1 - blend), tile) + np.multiply(blend, current_region)


def get_indicator_bound_point(indicator, lg, axis):
    '''Finds the index of first change in a binary mask 
    along a specified axis in a specified direction
    '''

    delta = np.diff(indicator, axis=axis)
    points = np.where(lg(delta, 0))
    del delta

    points = np.unique(points[axis])
    size = indicator.shape[axis]
    points = points[lg(points, size / 2.0)]
    
    if len(points) > 0:
        return points[-1]
    return None


def blend_component_from_point(point, mesh, lg):
    '''Obtains a normalized component of the blend, which describes depth of 
    overlap along a specified axis in a specified direction
    '''

    # this has the effect that the shallowest part of the blend 
    # is always 0 - symmetric with the deepest after normalization.
    blend = point - mesh + 1 
    blend[lg(blend, 0)] = 0

    blend = np.fabs(blend)
    mx = np.amax(blend)
    mx = mx if mx > 0.0 else 1.0

    return blend / mx


def get_blend_component(indicator, lg, axis, meshes):
    '''
    '''

    point = get_indicator_bound_point(indicator, lg, axis)
    if point is None:
        return []

    return [blend_component_from_point(point, meshes[axis], lg)]


def get_overall_blend(indicator, meshes):
    '''
    '''

    blends = []

    for lg in (op.lt, op.gt):
        for axis in (0, 1):
            blends.extend(get_blend_component(indicator, lg, axis, meshes))

    if len(blends) == 0:
        return np.zeros_like(indicator)
    return reduce(np.maximum, blends)


def get_blend(indicator_region, stup, cb=np.array):
    '''
    '''

    meshes = np.meshgrid(*map(np.arange, stup), indexing='ij')
    blend = get_overall_blend(indicator_region, meshes)

    return cb(np.multiply(blend, indicator_region))

        
